/* Copyright (C) 2020 Google, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#ifndef _ROGUE_PDS_ENCODE_H_
#define _ROGUE_PDS_ENCODE_H_

#include <assert.h>
#include <stdbool.h>
#include <stdint.h>
#include <util/bitset.h>
#include <util/log.h>



#define BITMASK_WORDS BITSET_WORDS(32)

typedef struct {
    BITSET_WORD bitset[BITMASK_WORDS];
} bitmask_t;

static inline uint64_t
bitmask_to_uint64_t(bitmask_t mask)
{
    return mask.bitset[0];
}

static inline bitmask_t
uint64_t_to_bitmask(uint64_t val)
{
    bitmask_t mask = {
        .bitset[0] = val & 0xffffffff,
    };

    return mask;
}

static inline void
store_instruction(BITSET_WORD *dst, bitmask_t instr)
{
    *(dst + 0) = instr.bitset[0];
}

/**
 * Opaque type from the PoV of generated code, but allows state to be passed
 * thru to the hand written helpers used by the generated code.
 */
struct encode_state;

/**
 * Allows to use gpu_id in expr functions
 */
#define ISA_GPU_ID() s->gen

struct bitset_params;

static bitmask_t
pack_field(unsigned low, unsigned high, int64_t val, bool is_signed)
{
   bitmask_t field, mask;

   if (is_signed) {
      /* NOTE: Don't assume val is already sign-extended to 64b,
       * just check that the bits above the valid range are either
       * all zero or all one:
       */
      assert(!(( val & ~BITFIELD64_MASK(1 + high - low)) &&
               (~val & ~BITFIELD64_MASK(1 + high - low))));
   } else {
      assert(!(val & ~BITFIELD64_MASK(1 + high - low)));
   }

   BITSET_ZERO(field.bitset);

   if (!val)
      return field;

   BITSET_ZERO(mask.bitset);
   BITSET_SET_RANGE(mask.bitset, 0, high - low);

   field = uint64_t_to_bitmask(val);
   BITSET_AND(field.bitset, field.bitset, mask.bitset);
   BITSET_SHL(field.bitset, low);

   return field;
}

/*
 * Forward-declarations (so we don't have to figure out which order to
 * emit various encoders when they have reference each other)
 */

static bitmask_t encode__reg_32(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src);
static bitmask_t encode__reg_tp_32(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src);
static bitmask_t encodereg_t_32(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src);
static bitmask_t encode__reg_64(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg  src);
static bitmask_t encode__reg_tp_64(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src);
static bitmask_t encodereg_t_64(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src);
static bitmask_t encodereg_c_64(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src);
static bitmask_t encode__instruction(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src);


/*
 * Expression evaluators:
 */

struct bitset_params {
};






static inline int64_t __reg_32_expr_anon_0(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src);
static inline int64_t __reg_32_expr_anon_1(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src);
static inline int64_t __reg_32_expr_anon_2(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src);
static inline int64_t __reg_tp_32_expr_anon_3(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src);
static inline int64_t __reg_tp_32_expr_anon_4(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src);
static inline int64_t reg_t_32_expr_anon_5(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src);
static inline int64_t __reg_64_expr_anon_6(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg  src);
static inline int64_t __reg_64_expr_anon_7(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg  src);
static inline int64_t __reg_64_expr_anon_8(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg  src);
static inline int64_t __reg_tp_64_expr_anon_9(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src);
static inline int64_t __reg_tp_64_expr_anon_10(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src);
static inline int64_t reg_t_64_expr_anon_11(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src);
static inline int64_t reg_c_64_expr_anon_12(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src);
static inline int64_t __instruction_expr_anon_13(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src);
static inline int64_t __instruction_expr_anon_14(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src);
static inline int64_t __instruction_expr_anon_15(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src);



          
static inline int64_t
__reg_32_expr_anon_0(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
    return PVR_PDS_REG_TYPE_CONST;
}


          
static inline int64_t
__reg_32_expr_anon_1(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
    return PVR_PDS_REG_TYPE_TEMP;
}


          
static inline int64_t
__reg_32_expr_anon_2(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
    return PVR_PDS_REG_TYPE_PTEMP;
}



          
static inline int64_t
__reg_tp_32_expr_anon_3(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
    return PVR_PDS_REG_TYPE_TEMP;
}


          
static inline int64_t
__reg_tp_32_expr_anon_4(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
    return PVR_PDS_REG_TYPE_PTEMP;
}



          
static inline int64_t
reg_t_32_expr_anon_5(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
    return PVR_PDS_REG_TYPE_TEMP;
}



          
static inline int64_t
__reg_64_expr_anon_6(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg  src)
{
    return PVR_PDS_REG_TYPE_CONST;
}


          
static inline int64_t
__reg_64_expr_anon_7(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg  src)
{
    return PVR_PDS_REG_TYPE_TEMP;
}


          
static inline int64_t
__reg_64_expr_anon_8(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg  src)
{
    return PVR_PDS_REG_TYPE_PTEMP;
}



          
static inline int64_t
__reg_tp_64_expr_anon_9(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
    return PVR_PDS_REG_TYPE_TEMP;
}


          
static inline int64_t
__reg_tp_64_expr_anon_10(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
    return PVR_PDS_REG_TYPE_PTEMP;
}



          
static inline int64_t
reg_t_64_expr_anon_11(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
    return PVR_PDS_REG_TYPE_TEMP;
}



          
static inline int64_t
reg_c_64_expr_anon_12(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
    return PVR_PDS_REG_TYPE_CONST;
}



          
static inline int64_t
__instruction_expr_anon_13(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
    int64_t IM;

          IM = src->srcs[2].type == PVR_PDS_SRC_TYPE_IMMEDIATE;
    return IM;
}


          
static inline int64_t
__instruction_expr_anon_14(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
    int64_t IM;

          IM = src->srcs[2].type == PVR_PDS_SRC_TYPE_IMMEDIATE;
    return IM;
}


          
static inline int64_t
__instruction_expr_anon_15(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
    int64_t IM;

          IM = (src->srcs[1].type == PVR_PDS_SRC_TYPE_IMMEDIATE);
    return IM == true;
}



/*
 * The actual encoder definitions
 */



static bitmask_t
snippet__reg_32_0(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src.idx;
             const bitmask_t packed = pack_field(0, 6, fld, false);  /* REG_IDX */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
snippet__reg_32_1(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src.idx;
             const bitmask_t packed = pack_field(0, 4, fld, false);  /* REG_IDX */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
encode__reg_32(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
   switch (__reg_32_case(s, src)) {
   case PVR_PDS_REG_32_CONST: {


      bitmask_t val = uint64_t_to_bitmask(0x0);




      BITSET_OR(val.bitset, val.bitset, snippet__reg_32_0(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_REG_32_TEMP: {


      bitmask_t val = uint64_t_to_bitmask(0x80);




      BITSET_OR(val.bitset, val.bitset, snippet__reg_32_1(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_REG_32_PTEMP: {


      bitmask_t val = uint64_t_to_bitmask(0xc0);




      BITSET_OR(val.bitset, val.bitset, snippet__reg_32_1(s, p, src).bitset);
      return val;
    }
   default:
      /* Note that we need the default case, because there are
       * instructions which we never expect to be encoded, (ie.
       * meta/macro instructions) as they are removed/replace
       * in earlier stages of the compiler.
       */
      break;
   }
   mesa_loge("Unhandled #reg-32 encode case: 0x%x\n", __reg_32_case(s, src));
   return uint64_t_to_bitmask(0);
}


static bitmask_t
snippet__reg_tp_32_0(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src.idx;
             const bitmask_t packed = pack_field(0, 4, fld, false);  /* REG_IDX */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
encode__reg_tp_32(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
   switch (__reg_tp_32_case(s, src)) {
   case PVR_PDS_REG_TP_32_TEMP: {


      bitmask_t val = uint64_t_to_bitmask(0x0);




      BITSET_OR(val.bitset, val.bitset, snippet__reg_tp_32_0(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_REG_TP_32_PTEMP: {


      bitmask_t val = uint64_t_to_bitmask(0x20);




      BITSET_OR(val.bitset, val.bitset, snippet__reg_tp_32_0(s, p, src).bitset);
      return val;
    }
   default:
      /* Note that we need the default case, because there are
       * instructions which we never expect to be encoded, (ie.
       * meta/macro instructions) as they are removed/replace
       * in earlier stages of the compiler.
       */
      break;
   }
   mesa_loge("Unhandled #reg-tp-32 encode case: 0x%x\n", __reg_tp_32_case(s, src));
   return uint64_t_to_bitmask(0);
}


static bitmask_t
snippetreg_t_32_0(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src.idx;
             const bitmask_t packed = pack_field(0, 4, fld, false);  /* REG_IDX */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}


static bitmask_t
encodereg_t_32(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
   switch (reg_t_32_case(s, src)) {
   case PVR_PDS_REG_T_32: {


      bitmask_t val = uint64_t_to_bitmask(0x0);




      BITSET_OR(val.bitset, val.bitset, snippetreg_t_32_0(s, p, src).bitset);
      return val;
    }
   default:
      /* Note that we need the default case, because there are
       * instructions which we never expect to be encoded, (ie.
       * meta/macro instructions) as they are removed/replace
       * in earlier stages of the compiler.
       */
      break;
   }
   mesa_loge("Unhandled reg-t-32 encode case: 0x%x\n", reg_t_32_case(s, src));
   return uint64_t_to_bitmask(0);
}


static bitmask_t
snippet__reg_64_0(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg  src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src.idx;
             const bitmask_t packed = pack_field(0, 5, fld, false);  /* REG_IDX */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
snippet__reg_64_1(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg  src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src.idx;
             const bitmask_t packed = pack_field(0, 4, fld, false);  /* REG_IDX */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
encode__reg_64(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg  src)
{
   switch (__reg_64_case(s, src)) {
   case PVR_PDS_REG_64_CONST: {


      bitmask_t val = uint64_t_to_bitmask(0x0);




      BITSET_OR(val.bitset, val.bitset, snippet__reg_64_0(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_REG_64_TEMP: {


      bitmask_t val = uint64_t_to_bitmask(0x40);




      BITSET_OR(val.bitset, val.bitset, snippet__reg_64_1(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_REG_64_PTEMP: {


      bitmask_t val = uint64_t_to_bitmask(0x60);




      BITSET_OR(val.bitset, val.bitset, snippet__reg_64_1(s, p, src).bitset);
      return val;
    }
   default:
      /* Note that we need the default case, because there are
       * instructions which we never expect to be encoded, (ie.
       * meta/macro instructions) as they are removed/replace
       * in earlier stages of the compiler.
       */
      break;
   }
   mesa_loge("Unhandled #reg-64 encode case: 0x%x\n", __reg_64_case(s, src));
   return uint64_t_to_bitmask(0);
}


static bitmask_t
snippet__reg_tp_64_0(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src.idx;
             const bitmask_t packed = pack_field(0, 3, fld, false);  /* REG_IDX */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
encode__reg_tp_64(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
   switch (__reg_tp_64_case(s, src)) {
   case PVR_PDS_REG_TP_64_TEMP: {


      bitmask_t val = uint64_t_to_bitmask(0x0);




      BITSET_OR(val.bitset, val.bitset, snippet__reg_tp_64_0(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_REG_TP_64_PTEMP: {


      bitmask_t val = uint64_t_to_bitmask(0x10);




      BITSET_OR(val.bitset, val.bitset, snippet__reg_tp_64_0(s, p, src).bitset);
      return val;
    }
   default:
      /* Note that we need the default case, because there are
       * instructions which we never expect to be encoded, (ie.
       * meta/macro instructions) as they are removed/replace
       * in earlier stages of the compiler.
       */
      break;
   }
   mesa_loge("Unhandled #reg-tp-64 encode case: 0x%x\n", __reg_tp_64_case(s, src));
   return uint64_t_to_bitmask(0);
}


static bitmask_t
snippetreg_t_64_0(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src.idx;
             const bitmask_t packed = pack_field(0, 3, fld, false);  /* REG_IDX */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}


static bitmask_t
encodereg_t_64(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
   switch (reg_t_64_case(s, src)) {
   case PVR_PDS_REG_T_64: {


      bitmask_t val = uint64_t_to_bitmask(0x0);




      BITSET_OR(val.bitset, val.bitset, snippetreg_t_64_0(s, p, src).bitset);
      return val;
    }
   default:
      /* Note that we need the default case, because there are
       * instructions which we never expect to be encoded, (ie.
       * meta/macro instructions) as they are removed/replace
       * in earlier stages of the compiler.
       */
      break;
   }
   mesa_loge("Unhandled reg-t-64 encode case: 0x%x\n", reg_t_64_case(s, src));
   return uint64_t_to_bitmask(0);
}


static bitmask_t
snippetreg_c_64_0(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src.idx;
             const bitmask_t packed = pack_field(0, 5, fld, false);  /* REG_IDX */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}


static bitmask_t
encodereg_c_64(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_reg src)
{
   switch (reg_c_64_case(s, src)) {
   case PVR_PDS_REG_C_64: {


      bitmask_t val = uint64_t_to_bitmask(0x0);




      BITSET_OR(val.bitset, val.bitset, snippetreg_c_64_0(s, p, src).bitset);
      return val;
    }
   default:
      /* Note that we need the default case, because there are
       * instructions which we never expect to be encoded, (ie.
       * meta/macro instructions) as they are removed/replace
       * in earlier stages of the compiler.
       */
      break;
   }
   mesa_loge("Unhandled reg-c-64 encode case: 0x%x\n", reg_c_64_case(s, src));
   return uint64_t_to_bitmask(0);
}


static bitmask_t
snippet__instruction_0(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src->pred.cc;
             const bitmask_t packed = pack_field(27, 27, fld, false);  /* CC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encodereg_t_64(s, &bp, src->srcs[PVR_PDS_INSTR_DST].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(0, 3, fld, false);  /* DST */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_32(s, &bp, src->srcs[0].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(19, 26, fld, false);  /* SRC0 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_32(s, &bp, src->srcs[1].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(11, 18, fld, false);  /* SRC1 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = !!(src->flags & PVR_PDS_INSTR_FLAG_INVERT);
             const bitmask_t packed = pack_field(29, 29, fld, false);  /* SNA */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = !!(src->flags & PVR_PDS_INSTR_FLAG_SIGNED_OP);
             const bitmask_t packed = pack_field(28, 28, fld, false);  /* ALUM */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_64(s, &bp, src->srcs[2].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(4, 10, fld, false);  /* SRC2 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
snippet__instruction_1(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    if (__instruction_expr_anon_13(s, p, src)) {


           
    {

             fld = src->pred.cc;
             const bitmask_t packed = pack_field(27, 27, fld, false);  /* CC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encodereg_t_32(s, &bp, src->srcs[PVR_PDS_INSTR_DST].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(0, 4, fld, false);  /* DST */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encodereg_t_32(s, &bp, src->srcs[0].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(19, 23, fld, false);  /* SRC0 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->sub_op_type.logic;
             const bitmask_t packed = pack_field(24, 26, fld, false);  /* LOP */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_32(s, &bp, src->srcs[1].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(11, 18, fld, false);  /* SRC1 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->srcs[2].imm;
             const bitmask_t packed = pack_field(5, 10, fld, false);  /* SRC2_IMM */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->srcs[2].type == PVR_PDS_SRC_TYPE_IMMEDIATE;
             const bitmask_t packed = pack_field(28, 28, fld, false);  /* IM */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    } else


    
    {


           
    {

             fld = src->pred.cc;
             const bitmask_t packed = pack_field(27, 27, fld, false);  /* CC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encodereg_t_32(s, &bp, src->srcs[PVR_PDS_INSTR_DST].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(0, 4, fld, false);  /* DST */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encodereg_t_32(s, &bp, src->srcs[0].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(19, 23, fld, false);  /* SRC0 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->sub_op_type.logic;
             const bitmask_t packed = pack_field(24, 26, fld, false);  /* LOP */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_32(s, &bp, src->srcs[1].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(11, 18, fld, false);  /* SRC1 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_tp_32(s, &bp, src->srcs[2].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(5, 10, fld, false);  /* SRC2 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->srcs[2].type == PVR_PDS_SRC_TYPE_IMMEDIATE;
             const bitmask_t packed = pack_field(28, 28, fld, false);  /* IM */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
snippet__instruction_2(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src->pred.so_global;
             const bitmask_t packed = pack_field(28, 28, fld, false);  /* CCS_GLOBAL */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->pred.so;
             const bitmask_t packed = pack_field(27, 27, fld, false);  /* CCS_SO */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->pred.cc;
             const bitmask_t packed = pack_field(26, 26, fld, false);  /* CCS_CC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->srcs[4].imm;
             const bitmask_t packed = pack_field(23, 24, fld, false);  /* SO */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_tp_64(s, &bp, src->srcs[0].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(18, 22, fld, false);  /* SRC0 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_tp_64(s, &bp, src->srcs[1].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(13, 17, fld, false);  /* SRC1 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_32(s, &bp, src->srcs[2].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(5, 12, fld, false);  /* SRC2 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_tp_64(s, &bp, src->srcs[3].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(0, 4, fld, false);  /* SRC3 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = !!(src->flags & PVR_PDS_INSTR_FLAG_TST);
             const bitmask_t packed = pack_field(25, 25, fld, false);  /* SO_TST */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
snippet__instruction_3(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src->pred.cc;
             const bitmask_t packed = pack_field(27, 27, fld, false);  /* CC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_tp_64(s, &bp, src->srcs[PVR_PDS_INSTR_DST].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(0, 4, fld, false);  /* DST */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_64(s, &bp, src->srcs[0].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(12, 18, fld, false);  /* SRC0 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = !!(src->flags & PVR_PDS_INSTR_FLAG_INVERT);
             const bitmask_t packed = pack_field(24, 24, fld, false);  /* SNA */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = !!(src->flags & PVR_PDS_INSTR_FLAG_SIGNED_OP);
             const bitmask_t packed = pack_field(26, 26, fld, false);  /* ALUM */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_64(s, &bp, src->srcs[1].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(5, 11, fld, false);  /* SRC1 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
snippet__instruction_4(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src->pred.cc;
             const bitmask_t packed = pack_field(27, 27, fld, false);  /* CC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_tp_32(s, &bp, src->srcs[PVR_PDS_INSTR_DST].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(0, 5, fld, false);  /* DST */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_32(s, &bp, src->srcs[0].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(14, 21, fld, false);  /* SRC0 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = !!(src->flags & PVR_PDS_INSTR_FLAG_INVERT);
             const bitmask_t packed = pack_field(24, 24, fld, false);  /* SNA */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = !!(src->flags & PVR_PDS_INSTR_FLAG_SIGNED_OP);
             const bitmask_t packed = pack_field(26, 26, fld, false);  /* ALUM */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_32(s, &bp, src->srcs[1].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(6, 13, fld, false);  /* SRC1 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
snippet__instruction_5(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    if (__instruction_expr_anon_14(s, p, src)) {


           
    {

             fld = src->pred.cc;
             const bitmask_t packed = pack_field(27, 27, fld, false);  /* CC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_tp_64(s, &bp, src->srcs[PVR_PDS_INSTR_DST].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(0, 4, fld, false);  /* DST */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_tp_64(s, &bp, src->srcs[0].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(18, 22, fld, false);  /* SRC0 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->sub_op_type.logic;
             const bitmask_t packed = pack_field(24, 26, fld, false);  /* LOP */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_tp_64(s, &bp, src->srcs[1].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(13, 17, fld, false);  /* SRC1 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->srcs[2].imm;
             const bitmask_t packed = pack_field(5, 10, fld, false);  /* SRC2_IMM */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->srcs[2].type == PVR_PDS_SRC_TYPE_IMMEDIATE;
             const bitmask_t packed = pack_field(23, 23, fld, false);  /* IM */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    } else


    
    {


           
    {

             fld = src->pred.cc;
             const bitmask_t packed = pack_field(27, 27, fld, false);  /* CC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_tp_64(s, &bp, src->srcs[PVR_PDS_INSTR_DST].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(0, 4, fld, false);  /* DST */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_tp_64(s, &bp, src->srcs[0].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(18, 22, fld, false);  /* SRC0 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->sub_op_type.logic;
             const bitmask_t packed = pack_field(24, 26, fld, false);  /* LOP */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_tp_64(s, &bp, src->srcs[1].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(13, 17, fld, false);  /* SRC1 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_32(s, &bp, src->srcs[2].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(5, 12, fld, false);  /* SRC2 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->srcs[2].type == PVR_PDS_SRC_TYPE_IMMEDIATE;
             const bitmask_t packed = pack_field(23, 23, fld, false);  /* IM */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
snippet__instruction_6(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    if (__instruction_expr_anon_15(s, p, src)) {


           
    {

             fld = src->pred.cc;
             const bitmask_t packed = pack_field(27, 27, fld, false);  /* CC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_tp_64(s, &bp, src->srcs[0].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(18, 22, fld, false);  /* SRC0 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->sub_op_type.comparison;
             const bitmask_t packed = pack_field(25, 26, fld, false);  /* COP */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->srcs[1].imm;
             const bitmask_t packed = pack_field(2, 17, fld, false);  /* SRC1_IMM */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = (src->srcs[1].type == PVR_PDS_SRC_TYPE_IMMEDIATE);
             const bitmask_t packed = pack_field(23, 23, fld, false);  /* IM */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    } else


    
    {


           
    {

             fld = src->pred.cc;
             const bitmask_t packed = pack_field(27, 27, fld, false);  /* CC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_tp_64(s, &bp, src->srcs[0].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(18, 22, fld, false);  /* SRC0 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->sub_op_type.comparison;
             const bitmask_t packed = pack_field(25, 26, fld, false);  /* COP */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_64(s, &bp, src->srcs[1].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(2, 8, fld, false);  /* SRC1 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = (src->srcs[1].type == PVR_PDS_SRC_TYPE_IMMEDIATE);
             const bitmask_t packed = pack_field(23, 23, fld, false);  /* IM */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
snippet__instruction_7(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src->srcs[0].imm;
             const bitmask_t packed = pack_field(24, 27, fld, false);  /* SRCC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = !!(src->flags & PVR_PDS_INSTR_FLAG_INVERT);
             const bitmask_t packed = pack_field(23, 23, fld, false);  /* NEG */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->srcs[2].imm;
             const bitmask_t packed = pack_field(0, 18, fld, false);  /* ADDR */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->srcs[1].imm;
             const bitmask_t packed = pack_field(19, 22, fld, false);  /* SETC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
snippet__instruction_8(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src->pred.cc;
             const bitmask_t packed = pack_field(27, 27, fld, false);  /* CC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_32(s, &bp, src->srcs[0].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(18, 25, fld, false);  /* SRC0 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encodereg_t_32(s, &bp, src->srcs[1].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(13, 17, fld, false);  /* SRC1 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_64(s, &bp, src->srcs[2].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(6, 12, fld, false);  /* SRC2 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = !!(src->flags & PVR_PDS_INSTR_FLAG_END);
             const bitmask_t packed = pack_field(26, 26, fld, false);  /* END */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
snippet__instruction_9(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src->pred.cc;
             const bitmask_t packed = pack_field(27, 27, fld, false);  /* CC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->sub_op_type.dout_cmd;
             const bitmask_t packed = pack_field(0, 2, fld, false);  /* CMD */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_64(s, &bp, src->srcs[0].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(8, 14, fld, false);  /* SRC0 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_32(s, &bp, src->srcs[1].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(16, 23, fld, false);  /* SRC1 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = !!(src->flags & PVR_PDS_INSTR_FLAG_END);
             const bitmask_t packed = pack_field(26, 26, fld, false);  /* END */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
snippet__instruction_10(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src->pred.cc;
             const bitmask_t packed = pack_field(27, 27, fld, false);  /* CC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encode__reg_64(s, &bp, src->srcs[0].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(0, 6, fld, false);  /* SRC0 */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}




static bitmask_t
snippet__instruction_11(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src->pred.cc;
             const bitmask_t packed = pack_field(27, 27, fld, false);  /* CC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
snippet__instruction_12(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src->pred.cc;
             const bitmask_t packed = pack_field(27, 27, fld, false);  /* CC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             { 
 struct bitset_params bp = {
 };

               bitmask_t tmp = encodereg_t_32(s, &bp, src->srcs[PVR_PDS_INSTR_DST].reg);
               fld = bitmask_to_uint64_t(tmp);
             }
             const bitmask_t packed = pack_field(18, 22, fld, false);  /* DST */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = !!(src->flags & PVR_PDS_INSTR_FLAG_IMM_IS_GR);
             const bitmask_t packed = pack_field(1, 1, fld, false);  /* GR */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->srcs[0].imm;
             const bitmask_t packed = pack_field(2, 17, fld, false);  /* SRC0_IMM */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}






static bitmask_t
snippet__instruction_13(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
   bitmask_t val = uint64_t_to_bitmask(0);









      uint64_t fld;

      (void)fld;


    
    {


           
    {

             fld = src->pred.cc;
             const bitmask_t packed = pack_field(27, 27, fld, false);  /* CC */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


           
    {

             fld = src->srcs[0].imm;
             const bitmask_t packed = pack_field(0, 4, fld, false);  /* SOMASK */
             BITSET_OR(val.bitset, val.bitset, packed.bitset);
             
    }


      {}  /* in case no unconditional field to close out last '} else' */
    
    }


   return val;
}



static bitmask_t
encode__instruction(struct encode_state *s, const struct bitset_params *p, const struct pvr_pds_instr * src)
{
   switch (__instruction_case(s, src)) {
   case PVR_PDS_OP_MAD: {


      bitmask_t val = uint64_t_to_bitmask(0x0);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_0(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_SFTLP32: {


      bitmask_t val = uint64_t_to_bitmask(0x40000000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_1(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_STM: {


      bitmask_t val = uint64_t_to_bitmask(0x60000000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_2(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_ADD64: {


      bitmask_t val = uint64_t_to_bitmask(0x80000000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_3(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_ADD32: {


      bitmask_t val = uint64_t_to_bitmask(0x90000000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_4(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_SFTLP64: {


      bitmask_t val = uint64_t_to_bitmask(0xa0000000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_5(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_CMP: {


      bitmask_t val = uint64_t_to_bitmask(0xb1000000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_6(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_BRA: {


      bitmask_t val = uint64_t_to_bitmask(0xc0000000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_7(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_DDMAD: {


      bitmask_t val = uint64_t_to_bitmask(0xe0000000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_8(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_DOUT: {


      bitmask_t val = uint64_t_to_bitmask(0xf0000000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_9(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_LD: {


      bitmask_t val = uint64_t_to_bitmask(0xd0000000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_10(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_ST: {


      bitmask_t val = uint64_t_to_bitmask(0xd0800000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_10(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_WDF: {


      bitmask_t val = uint64_t_to_bitmask(0xd1000000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_11(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_LIMM: {


      bitmask_t val = uint64_t_to_bitmask(0xd1800000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_12(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_LOCK: {


      bitmask_t val = uint64_t_to_bitmask(0xd2000000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_11(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_RELEASE: {


      bitmask_t val = uint64_t_to_bitmask(0xd2800000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_11(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_HALT: {


      bitmask_t val = uint64_t_to_bitmask(0xd3000000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_11(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_STMC: {


      bitmask_t val = uint64_t_to_bitmask(0xd3800000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_13(s, p, src).bitset);
      return val;
    }
   case PVR_PDS_OP_NOP: {


      bitmask_t val = uint64_t_to_bitmask(0xd7800000);




      BITSET_OR(val.bitset, val.bitset, snippet__instruction_11(s, p, src).bitset);
      return val;
    }
   default:
      /* Note that we need the default case, because there are
       * instructions which we never expect to be encoded, (ie.
       * meta/macro instructions) as they are removed/replace
       * in earlier stages of the compiler.
       */
      break;
   }
   mesa_loge("Unhandled #instruction encode case: 0x%x\n", __instruction_case(s, src));
   return uint64_t_to_bitmask(0);
}

#endif /* _ROGUE_PDS_ENCODE_H_ */
