/*
 * Copyright © 2024 Imagination Technologies Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdbool.h>
#include <stdint.h>

#include "pvr_pds_isa_encode.h"
#include "util/macros.h"

/* Define things before including the isaspec generated encoder header. */

struct encode_state;
struct pvr_pds_instr;
struct pvr_pds_reg;

enum pvr_pds_reg_mangled_name {
   PVR_PDS_REG_32_CONST,
   PVR_PDS_REG_32_TEMP,
   PVR_PDS_REG_32_PTEMP,
   PVR_PDS_REG_TP_32_TEMP,
   PVR_PDS_REG_TP_32_PTEMP,
   PVR_PDS_REG_T_32,
   PVR_PDS_REG_64_CONST,
   PVR_PDS_REG_64_TEMP,
   PVR_PDS_REG_64_PTEMP,
   PVR_PDS_REG_TP_64_TEMP,
   PVR_PDS_REG_TP_64_PTEMP,
   PVR_PDS_REG_T_64,
   PVR_PDS_REG_C_64,
};

static enum pvr_pds_reg_mangled_name __reg_32_case(struct encode_state *s,
                                                   struct pvr_pds_reg src);

static enum pvr_pds_reg_mangled_name __reg_tp_32_case(struct encode_state *s,
                                                      struct pvr_pds_reg src);

static enum pvr_pds_reg_mangled_name reg_t_32_case(struct encode_state *s,
                                                   struct pvr_pds_reg src);

static enum pvr_pds_reg_mangled_name __reg_64_case(struct encode_state *s,
                                                   struct pvr_pds_reg src);

static enum pvr_pds_reg_mangled_name __reg_tp_64_case(struct encode_state *s,
                                                      struct pvr_pds_reg src);

static enum pvr_pds_reg_mangled_name reg_t_64_case(struct encode_state *s,
                                                   struct pvr_pds_reg src);

static enum pvr_pds_reg_mangled_name reg_c_64_case(struct encode_state *s,
                                                   struct pvr_pds_reg src);

static enum pvr_pds_op_type __instruction_case(struct encode_state *s,
                                               struct pvr_pds_instr *src);

#include "pvr_pds_isa.h"
#include "rogue_pds_encode.h"

static enum pvr_pds_reg_mangled_name __reg_32_case(struct encode_state *s,
                                                   struct pvr_pds_reg src)
{
   assert(src.reg_size == PVR_PDS_REG_SIZE_32);

   switch (src.type) {
   case PVR_PDS_REG_TYPE_CONST:
      return PVR_PDS_REG_32_CONST;
   case PVR_PDS_REG_TYPE_TEMP:
      return PVR_PDS_REG_32_TEMP;
   case PVR_PDS_REG_TYPE_PTEMP:
      return PVR_PDS_REG_32_PTEMP;
   default:
      unreachable("Invalid/unknown reg type");
   }
}

static enum pvr_pds_reg_mangled_name __reg_tp_32_case(struct encode_state *s,
                                                      struct pvr_pds_reg src)
{
   assert(src.reg_size == PVR_PDS_REG_SIZE_32);

   switch (src.type) {
   case PVR_PDS_REG_TYPE_CONST:
      unreachable("CONST reg type is not supported for #reg-tp-32");
   case PVR_PDS_REG_TYPE_TEMP:
      return PVR_PDS_REG_TP_32_TEMP;
   case PVR_PDS_REG_TYPE_PTEMP:
      return PVR_PDS_REG_TP_32_PTEMP;
   default:
      unreachable("Invalid/unknown reg type");
   }
}

static enum pvr_pds_reg_mangled_name reg_t_32_case(struct encode_state *s,
                                                   struct pvr_pds_reg src)
{
   assert(src.type == PVR_PDS_REG_TYPE_TEMP);
   assert(src.reg_size == PVR_PDS_REG_SIZE_32);
   return PVR_PDS_REG_T_32;
}

static enum pvr_pds_reg_mangled_name __reg_64_case(struct encode_state *s,
                                                   struct pvr_pds_reg src)
{
   assert(src.reg_size == PVR_PDS_REG_SIZE_64);

   switch (src.type) {
   case PVR_PDS_REG_TYPE_CONST:
      return PVR_PDS_REG_64_CONST;
   case PVR_PDS_REG_TYPE_TEMP:
      return PVR_PDS_REG_64_TEMP;
   case PVR_PDS_REG_TYPE_PTEMP:
      return PVR_PDS_REG_64_PTEMP;
   default:
      unreachable("Invalid/unknown reg type");
   }
}

static enum pvr_pds_reg_mangled_name __reg_tp_64_case(struct encode_state *s,
                                                      struct pvr_pds_reg src)
{
   assert(src.reg_size == PVR_PDS_REG_SIZE_64);

   switch (src.type) {
   case PVR_PDS_REG_TYPE_CONST:
      unreachable("CONST reg type is not supported for #reg-tp-32");
   case PVR_PDS_REG_TYPE_TEMP:
      return PVR_PDS_REG_TP_64_TEMP;
   case PVR_PDS_REG_TYPE_PTEMP:
      return PVR_PDS_REG_TP_64_PTEMP;
   default:
      unreachable("Invalid/unknown reg type");
   }
}

static enum pvr_pds_reg_mangled_name reg_t_64_case(struct encode_state *s,
                                                   struct pvr_pds_reg src)
{
   assert(src.type == PVR_PDS_REG_TYPE_TEMP);
   assert(src.reg_size == PVR_PDS_REG_SIZE_64);
   return PVR_PDS_REG_T_64;
}

static enum pvr_pds_reg_mangled_name reg_c_64_case(struct encode_state *s,
                                                   struct pvr_pds_reg src)
{
   assert(src.type == PVR_PDS_REG_TYPE_CONST);
   assert(src.reg_size == PVR_PDS_REG_SIZE_64);
   return PVR_PDS_REG_C_64;
}

static enum pvr_pds_op_type __instruction_case(struct encode_state *s,
                                               struct pvr_pds_instr *src)
{
   return src->type;
}

size_t pvr_pds_encode_instr(struct pvr_pds_instr *instr, void *output_buffer)
{
   *(bitmask_t *)output_buffer = encode__instruction(NULL, NULL, instr);
   return sizeof(bitmask_t);
}
