/*
 * Copyright © 2024 Imagination Technologies Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include "nir/nir.h"
#include "nir/nir_builder.h"
#include "rogue.h"
#include "util/macros.h"

#include <assert.h>
#include <stdbool.h>
#include <stdint.h>

/**
 * \file rogue_nir_lower_alpha.c
 *
 * \brief Contains alpha to coverage lowering pass.
 */

static bool
lower_a2c(struct nir_builder *b, nir_instr *instr, void *data)
{
   if (instr->type != nir_instr_type_intrinsic)
      return false;

   nir_intrinsic_instr *intr = nir_instr_as_intrinsic(instr);
   if (intr->intrinsic != nir_intrinsic_store_output)
      return false;

   nir_io_semantics io_sem = nir_intrinsic_io_semantics(intr);
   if (io_sem.location < FRAG_RESULT_DATA0 ||
       io_sem.location > FRAG_RESULT_DATA7)
      return false;

   b->cursor = nir_before_instr(instr);

   nir_def *value = intr->src[0].ssa;
   nir_def *alpha_value = value->num_components < 4 ? nir_imm_float(b, 1.0f) : nir_channel(b, value, 3);

   nir_def *alpha_cov_msk = nir_ishl(b, nir_imm_int(b, 1), nir_load_sample_id(b));
   alpha_cov_msk = nir_iand(b, alpha_cov_msk, nir_alpha_to_coverage_img(b, alpha_value));
   alpha_cov_msk = nir_iand(b, alpha_cov_msk, nir_load_savmsk_vm_img(b)); /* TODO NEXT: do this properly... */
   nir_discard_if(b, nir_ieq_imm(b, alpha_cov_msk, 0));

   return true;
}

PUBLIC
bool rogue_nir_lower_a2c(nir_shader *shader, rogue_build_ctx *ctx)
{
   if (!ctx->stage_data.fs.alpha_to_coverage_enable)
      return false;

   return nir_shader_instructions_pass(shader, lower_a2c, nir_metadata_block_index | nir_metadata_dominance, NULL);
}
