/*
 * Copyright © 2024 Imagination Technologies Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <assert.h>
#include <stdint.h>
#include <string.h>

#include "pds/pvr_pds_isa.h"
#include "pvr_device_info.h"
#include "pvr_pds_builder.h"

#define pvr_pds_reg_size(_instr_src)                      \
   ({                                                     \
      uint32_t reg_size = PVR_PDS_REG_SIZE_INVALID;       \
      if (pvr_instr_src_length(_instr_src) == 1) {        \
         reg_size = PVR_PDS_REG_SIZE_32;                  \
      } else if (pvr_instr_src_length(_instr_src) == 2) { \
         reg_size = PVR_PDS_REG_SIZE_64;                  \
      } else {                                            \
         unreachable("Unknown reg size.");                \
      }                                                   \
      reg_size;                                           \
   })

static struct pvr_pds_instr_src
pvr_pds_instr_src_from_reg(const struct pvr_pds_reg reg)
{
   return (struct pvr_pds_instr_src){
      .type = PVR_PDS_SRC_TYPE_REGISTER,
      .reg = reg,
   };
}

static struct pvr_pds_instr_src
pvr_pds_build_instr_src_reg(enum pvr_pds_reg_type reg_type,
                            enum pvr_pds_reg_size reg_size,
                            uint32_t reg_idx)
{
   struct pvr_pds_reg reg = {
      .type = reg_type,
      .reg_size = reg_size,
      .idx = reg_idx,
   };

   return pvr_pds_instr_src_from_reg(reg);
}

static struct pvr_pds_instr_src pvr_pds_instr_src_from_imm(uint32_t imm)
{
   return (struct pvr_pds_instr_src){
      .type = PVR_PDS_SRC_TYPE_IMMEDIATE,
      .imm = imm,
   };
}

void pvr_pds_build_mad(struct pvr_pds_instr *instr,
                       const struct pvr_pds_reg dst,
                       const struct pvr_pds_reg mul0,
                       const struct pvr_pds_reg mul1,
                       const struct pvr_pds_reg add,
                       uint32_t flags)
{
   ASSERTED const uint32_t valid_flags = PVR_PDS_INSTR_FLAG_INVERT |
                                         PVR_PDS_INSTR_FLAG_SIGNED_OP;

   *instr = (struct pvr_pds_instr){
      .type = PVR_PDS_OP_MAD,
   };

   assert((flags & ~valid_flags) == 0);
   instr->flags = flags;

   assert(mul0.type != PVR_PDS_REG_TYPE_INVALID);
   assert(mul0.reg_size == PVR_PDS_REG_SIZE_32);
   instr->srcs[0] = pvr_pds_instr_src_from_reg(mul0);

   assert(mul1.type != PVR_PDS_REG_TYPE_INVALID);
   assert(mul1.reg_size == PVR_PDS_REG_SIZE_32);
   instr->srcs[1] = pvr_pds_instr_src_from_reg(mul1);

   assert(add.type != PVR_PDS_REG_TYPE_INVALID);
   assert(add.reg_size == PVR_PDS_REG_SIZE_64);
   instr->srcs[2] = pvr_pds_instr_src_from_reg(add);

   assert(dst.type == PVR_PDS_REG_TYPE_TEMP);
   assert(dst.reg_size == PVR_PDS_REG_SIZE_64);
   instr->srcs[PVR_PDS_INSTR_DST] = pvr_pds_instr_src_from_reg(dst);
}

static void pvr_pds_build_sftlp32_internal(struct pvr_pds_instr *instr,
                                           enum pvr_pds_logic_op_type logic_op,
                                           const struct pvr_pds_reg dst,
                                           const struct pvr_pds_reg src0,
                                           const struct pvr_pds_reg src1)
{
   *instr = (struct pvr_pds_instr) {
      .type = PVR_PDS_OP_SFTLP32,
      .sub_op_type = {
         .logic = logic_op,
      },
   };

   assert(src0.type == PVR_PDS_REG_TYPE_TEMP);
   assert(src0.reg_size == PVR_PDS_REG_SIZE_32);
   instr->srcs[0] = pvr_pds_instr_src_from_reg(src0);

   assert(src1.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src1.reg_size == PVR_PDS_REG_SIZE_32);
   instr->srcs[1] = pvr_pds_instr_src_from_reg(src1);

   assert(dst.type == PVR_PDS_REG_TYPE_TEMP);
   assert(dst.reg_size == PVR_PDS_REG_SIZE_32);
   instr->srcs[PVR_PDS_INSTR_DST] = pvr_pds_instr_src_from_reg(dst);
}

void pvr_pds_build_sftlp32(struct pvr_pds_instr *instr,
                           enum pvr_pds_logic_op_type logic_op,
                           const struct pvr_pds_reg dst,
                           const struct pvr_pds_reg src0,
                           const struct pvr_pds_reg src1,
                           const struct pvr_pds_reg shift)
{
   pvr_pds_build_sftlp32_internal(instr, logic_op, dst, src0, src1);

   assert(shift.type == PVR_PDS_REG_TYPE_TEMP ||
          shift.type == PVR_PDS_REG_TYPE_PTEMP);
   assert(shift.reg_size == PVR_PDS_REG_SIZE_32);

   instr->srcs[2] = pvr_pds_instr_src_from_reg(shift);
}

void pvr_pds_build_sftlp32i(struct pvr_pds_instr *instr,
                            enum pvr_pds_logic_op_type logic_op,
                            const struct pvr_pds_reg dst,
                            const struct pvr_pds_reg src0,
                            const struct pvr_pds_reg src1,
                            int32_t shift)
{
   uint32_t imm;

   pvr_pds_build_sftlp32_internal(instr, logic_op, dst, src0, src1);

   assert(shift == CLAMP(shift, -31, 31));

   imm = shift;
   imm &= 0b111111;

   instr->srcs[2] = pvr_pds_instr_src_from_imm(imm);
}

static void pvr_pds_build_sftlp64_internal(struct pvr_pds_instr *instr,
                                           enum pvr_pds_logic_op_type logic_op,
                                           const struct pvr_pds_reg dst,
                                           const struct pvr_pds_reg src0,
                                           const struct pvr_pds_reg src1)
{
   *instr = (struct pvr_pds_instr) {
      .type = PVR_PDS_OP_SFTLP64,
      .sub_op_type = {
         .logic = logic_op,
      },
   };

   assert(src0.type == PVR_PDS_REG_TYPE_TEMP ||
          src0.type == PVR_PDS_REG_TYPE_PTEMP);
   assert(src0.reg_size == PVR_PDS_REG_SIZE_64);
   instr->srcs[0] = pvr_pds_instr_src_from_reg(src0);

   assert(src1.type == PVR_PDS_REG_TYPE_TEMP ||
          src1.type == PVR_PDS_REG_TYPE_PTEMP);
   assert(src1.reg_size == PVR_PDS_REG_SIZE_64);
   instr->srcs[1] = pvr_pds_instr_src_from_reg(src1);

   assert(dst.type == PVR_PDS_REG_TYPE_TEMP ||
          dst.type == PVR_PDS_REG_TYPE_PTEMP);
   assert(dst.reg_size == PVR_PDS_REG_SIZE_64);
   instr->srcs[PVR_PDS_INSTR_DST] = pvr_pds_instr_src_from_reg(dst);
}

void pvr_pds_build_sftlp64(struct pvr_pds_instr *instr,
                           enum pvr_pds_logic_op_type logic_op,
                           const struct pvr_pds_reg dst,
                           const struct pvr_pds_reg src0,
                           const struct pvr_pds_reg src1,
                           const struct pvr_pds_reg shift)
{
   pvr_pds_build_sftlp64_internal(instr, logic_op, dst, src0, src1);

   assert(shift.type != PVR_PDS_REG_TYPE_INVALID);
   assert(shift.reg_size == PVR_PDS_REG_SIZE_32);

   instr->srcs[2] = pvr_pds_instr_src_from_reg(shift);
}

void pvr_pds_build_sftlp64i(struct pvr_pds_instr *instr,
                            enum pvr_pds_logic_op_type logic_op,
                            const struct pvr_pds_reg dst,
                            const struct pvr_pds_reg src0,
                            const struct pvr_pds_reg src1,
                            int32_t shift)
{
   uint32_t imm;

   pvr_pds_build_sftlp64_internal(instr, logic_op, dst, src0, src1);

   assert(shift == CLAMP(shift, -63, 63));

   imm = shift;
   imm &= 0b1111111;

   instr->srcs[2] = pvr_pds_instr_src_from_imm(imm);
}

void pvr_pds_build_add32(ASSERTED const struct pvr_device_info *dev_info,
                         struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg dst,
                         const struct pvr_pds_reg src0,
                         const struct pvr_pds_reg src1,
                         uint32_t flags)
{
   ASSERTED const uint32_t valid_flags = PVR_PDS_INSTR_FLAG_INVERT |
                                         PVR_PDS_INSTR_FLAG_SIGNED_OP;

   *instr = (struct pvr_pds_instr){
      .type = PVR_PDS_OP_ADD32,
   };

   assert((flags & ~valid_flags) == 0);
   instr->flags = flags;

   assert(src0.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src0.reg_size == PVR_PDS_REG_SIZE_32);
   /* With 37830 src0 has to be a reg aligned for 64 bit addressing. */
   assert(!PVR_HAS_QUIRK(dev_info, 37830) || (src0.idx & 1) == 0);
   instr->srcs[0] = pvr_pds_instr_src_from_reg(src0);

   assert(src1.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src1.reg_size == PVR_PDS_REG_SIZE_32);
   instr->srcs[1] = pvr_pds_instr_src_from_reg(src1);

   assert(dst.type == PVR_PDS_REG_TYPE_TEMP ||
          dst.type == PVR_PDS_REG_TYPE_PTEMP);
   assert(dst.reg_size == PVR_PDS_REG_SIZE_32);
   instr->srcs[PVR_PDS_INSTR_DST] = pvr_pds_instr_src_from_reg(dst);
}

void pvr_pds_build_add64(struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg dst,
                         const struct pvr_pds_reg src0,
                         const struct pvr_pds_reg src1,
                         uint32_t flags)
{
   ASSERTED const uint32_t valid_flags = PVR_PDS_INSTR_FLAG_INVERT |
                                         PVR_PDS_INSTR_FLAG_SIGNED_OP;

   *instr = (struct pvr_pds_instr){
      .type = PVR_PDS_OP_ADD64,
   };

   assert((flags & ~valid_flags) == 0);
   instr->flags = flags;

   assert(src0.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src0.reg_size == PVR_PDS_REG_SIZE_64);
   instr->srcs[0] = pvr_pds_instr_src_from_reg(src0);

   assert(src1.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src1.reg_size == PVR_PDS_REG_SIZE_64);
   instr->srcs[1] = pvr_pds_instr_src_from_reg(src1);

   assert(dst.type == PVR_PDS_REG_TYPE_TEMP ||
          dst.type == PVR_PDS_REG_TYPE_PTEMP);
   assert(dst.reg_size == PVR_PDS_REG_SIZE_64);
   instr->srcs[PVR_PDS_INSTR_DST] = pvr_pds_instr_src_from_reg(dst);
}

void pvr_pds_build_stm(struct pvr_pds_instr *instr,
                       uint32_t so_idx,
                       const struct pvr_pds_reg src0,
                       const struct pvr_pds_reg src1,
                       const struct pvr_pds_reg src2,
                       const struct pvr_pds_reg src3,
                       uint32_t flags)
{
   *instr = (struct pvr_pds_instr){
      .type = PVR_PDS_OP_STM,
   };

   assert(flags == 0 || flags == PVR_PDS_INSTR_FLAG_TST);
   instr->flags = flags;

   assert(src0.type == PVR_PDS_REG_TYPE_TEMP ||
          src0.type == PVR_PDS_REG_TYPE_PTEMP);
   assert(src0.reg_size == pvr_pds_reg_size(SO_SRC0));
   instr->srcs[0] = pvr_pds_instr_src_from_reg(src0);

   assert(src1.type == PVR_PDS_REG_TYPE_TEMP ||
          src1.type == PVR_PDS_REG_TYPE_PTEMP);
   assert(src1.reg_size == pvr_pds_reg_size(SO_SRC1));
   instr->srcs[1] = pvr_pds_instr_src_from_reg(src1);

   assert(src2.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src2.reg_size == pvr_pds_reg_size(SO_SRC2));
   instr->srcs[2] = pvr_pds_instr_src_from_reg(src2);

   assert(src3.type == PVR_PDS_REG_TYPE_TEMP ||
          src3.type == PVR_PDS_REG_TYPE_PTEMP);
   assert(src3.reg_size == pvr_pds_reg_size(SO_SRC3));
   instr->srcs[3] = pvr_pds_instr_src_from_reg(src3);

   instr->srcs[4] = pvr_pds_instr_src_from_imm(so_idx);
}

void pvr_pds_build_stmc(struct pvr_pds_instr *instr, uint32_t so_mask)
{
   ASSERTED const uint32_t valid_mask =
      PVR_PDS_SO_MASK_STREAM_0 | PVR_PDS_SO_MASK_STREAM_1 |
      PVR_PDS_SO_MASK_STREAM_2 | PVR_PDS_SO_MASK_STREAM_3 |
      PVR_PDS_SO_MASK_GLOBAL;

   *instr = (struct pvr_pds_instr){
      .type = PVR_PDS_OP_STMC,
   };

   assert((so_mask & ~valid_mask) == 0);
   instr->srcs[0] = pvr_pds_instr_src_from_imm(so_mask);
}

void pvr_pds_build_cmp(struct pvr_pds_instr *instr,
                       enum pvr_pds_comparison_op_type compare_op,
                       const struct pvr_pds_reg src0,
                       const struct pvr_pds_reg src1)
{
   *instr = (struct pvr_pds_instr) {
      .type = PVR_PDS_OP_CMP,
      .sub_op_type = {
         .comparison = compare_op,
      },
   };

   assert(src0.type == PVR_PDS_REG_TYPE_TEMP ||
          src0.type == PVR_PDS_REG_TYPE_PTEMP);
   assert(src0.reg_size == PVR_PDS_REG_SIZE_64);
   instr->srcs[0] = pvr_pds_instr_src_from_reg(src0);

   assert(src1.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src1.reg_size == PVR_PDS_REG_SIZE_64);
   instr->srcs[1] = pvr_pds_instr_src_from_reg(src1);
}

void pvr_pds_build_bra(struct pvr_pds_instr *instr,
                       enum pvr_pds_predicate src_pred,
                       enum pvr_pds_predicate set_pred,
                       uint32_t branch_offset,
                       uint32_t flags)
{
   *instr = (struct pvr_pds_instr){
      .type = PVR_PDS_OP_BRA,
   };

   assert(flags == 0 || flags == PVR_PDS_INSTR_FLAG_INVERT);
   instr->flags = flags;

   instr->srcs[0] = pvr_pds_instr_src_from_imm(src_pred);
   instr->srcs[1] = pvr_pds_instr_src_from_imm(set_pred);
   instr->srcs[2] = pvr_pds_instr_src_from_imm(branch_offset);
}

void pvr_pds_build_ddmad(struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg src0,
                         const struct pvr_pds_reg src1,
                         const struct pvr_pds_reg src2,
                         const struct pvr_pds_reg src3,
                         uint32_t flags)
{
   *instr = (struct pvr_pds_instr){
      .type = PVR_PDS_OP_DDMAD,
   };

   assert(flags == 0 || flags == PVR_PDS_INSTR_FLAG_END);
   instr->flags = flags;

   assert(src0.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src0.reg_size == pvr_pds_reg_size(DDMAD_SRC0));
   instr->srcs[0] = pvr_pds_instr_src_from_reg(src0);

   assert(src1.type == PVR_PDS_REG_TYPE_TEMP);
   assert(src1.reg_size == pvr_pds_reg_size(DDMAD_SRC1));
   instr->srcs[1] = pvr_pds_instr_src_from_reg(src1);

   assert(src2.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src2.reg_size == pvr_pds_reg_size(DDMAD_SRC2));
   instr->srcs[2] = pvr_pds_instr_src_from_reg(src2);

   assert(src3.type == PVR_PDS_REG_TYPE_CONST);
   assert(src3.reg_size == pvr_pds_reg_size(DDMAD_SRC3));
   instr->srcs[3] = pvr_pds_instr_src_from_reg(src3);
}

void pvr_pds_build_doutu(struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg src0,
                         uint32_t flags)
{
   *instr = (struct pvr_pds_instr) {
      .type = PVR_PDS_OP_DOUT,
      .sub_op_type = {
         .dout_cmd = PVR_PDS_DOUT_CMD_DOUTU,
      },
   };

   assert(flags == 0 || flags == PVR_PDS_INSTR_FLAG_END);
   instr->flags = flags;

   assert(src0.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src0.reg_size == pvr_pds_reg_size(DOUTU_SRC0));
   instr->srcs[0] = pvr_pds_instr_src_from_reg(src0);

   /* DOUTU_SRC1 is unused so it needs just be a valid register. We can be sure
    * that reg 0 is allocated since there's at least one instruction in the
    * program, this DOUTU for which there would have been an allocation for
    * DOUTU_SRC0.
    */
   instr->srcs[1] = pvr_pds_build_instr_src_reg(PVR_PDS_REG_TYPE_CONST,
                                                pvr_pds_reg_size(DOUTU_SRC1),
                                                0);
}

void pvr_pds_build_doutd(struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg src0,
                         const struct pvr_pds_reg src1,
                         uint32_t flags)
{
   *instr = (struct pvr_pds_instr) {
      .type = PVR_PDS_OP_DOUT,
      .sub_op_type = {
         .dout_cmd = PVR_PDS_DOUT_CMD_DOUTD,
      },
   };

   assert(flags == 0 || flags == PVR_PDS_INSTR_FLAG_END);
   instr->flags = flags;

   assert(src0.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src0.reg_size == pvr_pds_reg_size(DOUTD_SRC0));
   instr->srcs[0] = pvr_pds_instr_src_from_reg(src0);

   assert(src1.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src1.reg_size == pvr_pds_reg_size(DOUTD_SRC1));
   instr->srcs[1] = pvr_pds_instr_src_from_reg(src1);
}

void pvr_pds_build_doutw(struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg src0,
                         const struct pvr_pds_reg src1,
                         uint32_t flags)
{
   *instr = (struct pvr_pds_instr) {
      .type = PVR_PDS_OP_DOUT,
      .sub_op_type = {
         .dout_cmd = PVR_PDS_DOUT_CMD_DOUTW,
      },
   };

   assert(flags == 0 || flags == PVR_PDS_INSTR_FLAG_END);
   instr->flags = flags;

   assert(src0.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src0.reg_size == pvr_pds_reg_size(DOUTW_SRC0));
   instr->srcs[0] = pvr_pds_instr_src_from_reg(src0);

   assert(src1.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src1.reg_size == pvr_pds_reg_size(DOUTW_SRC1));
   instr->srcs[1] = pvr_pds_instr_src_from_reg(src1);
}

void pvr_pds_build_doutv(struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg src0,
                         const struct pvr_pds_reg src1,
                         uint32_t flags)
{
   *instr = (struct pvr_pds_instr) {
      .type = PVR_PDS_OP_DOUT,
      .sub_op_type = {
         .dout_cmd = PVR_PDS_DOUT_CMD_DOUTV,
      },
   };

   assert(flags == 0 || flags == PVR_PDS_INSTR_FLAG_END);
   instr->flags = flags;

   assert(src0.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src0.reg_size == pvr_pds_reg_size(DOUTV_SRC0));
   instr->srcs[0] = pvr_pds_instr_src_from_reg(src0);

   assert(src1.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src1.reg_size == pvr_pds_reg_size(DOUTV_SRC1));
   instr->srcs[1] = pvr_pds_instr_src_from_reg(src1);
}

void pvr_pds_build_doutc(struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg src0,
                         const struct pvr_pds_reg src1,
                         uint32_t flags)
{
   *instr = (struct pvr_pds_instr) {
      .type = PVR_PDS_OP_DOUT,
      .sub_op_type = {
         .dout_cmd = PVR_PDS_DOUT_CMD_DOUTC,
      },
   };

   assert(flags == 0 || flags == PVR_PDS_INSTR_FLAG_END);
   instr->flags = flags;

   assert(src0.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src0.reg_size == pvr_pds_reg_size(DOUTC_SRC0));
   instr->srcs[0] = pvr_pds_instr_src_from_reg(src0);

   assert(src1.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src1.reg_size == pvr_pds_reg_size(DOUTC_SRC1));
   instr->srcs[1] = pvr_pds_instr_src_from_reg(src1);
}

void pvr_pds_build_douti(struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg src0,
                         const struct pvr_pds_reg src1,
                         uint32_t flags)
{
   *instr = (struct pvr_pds_instr) {
      .type = PVR_PDS_OP_DOUT,
      .sub_op_type = {
         .dout_cmd = PVR_PDS_DOUT_CMD_DOUTI,
      },
   };

   assert(flags == 0 || flags == PVR_PDS_INSTR_FLAG_END);
   instr->flags = flags;

   assert(src0.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src0.reg_size == pvr_pds_reg_size(DOUTI_SRC0));
   instr->srcs[0] = pvr_pds_instr_src_from_reg(src0);

   assert(src1.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src1.reg_size == pvr_pds_reg_size(DOUTI_SRC1));
   instr->srcs[1] = pvr_pds_instr_src_from_reg(src1);
}

void pvr_pds_build_limm(struct pvr_pds_instr *instr,
                        const struct pvr_pds_reg dst,
                        uint16_t imm16)
{
   *instr = (struct pvr_pds_instr){
      .type = PVR_PDS_OP_LIMM,
   };

   instr->srcs[0] = pvr_pds_instr_src_from_imm(imm16);

   assert(dst.type == PVR_PDS_REG_TYPE_TEMP);
   assert(dst.reg_size == PVR_PDS_REG_SIZE_32);
   instr->srcs[PVR_PDS_INSTR_DST] = pvr_pds_instr_src_from_reg(dst);
}

void pvr_pds_build_limm64(struct pvr_pds_instr *instr,
                          const struct pvr_pds_reg dst,
                          uint64_t imm64)
{
   // Split the 64-bit immediate value into two 32-bit parts
   uint32_t imm_low = (uint32_t)(imm64 & 0xFFFFFFFF);
   uint32_t imm_high = (uint32_t)((imm64 >> 32) & 0xFFFFFFFF);

   // Ensure the destination register can accommodate 64-bit value
   assert(dst.type == PVR_PDS_REG_TYPE_TEMP);
   assert(dst.reg_size == PVR_PDS_REG_SIZE_64);

   pvr_pds_build_limm(instr, pvr_pds_ra_low_32b_from_64b(dst), imm_low);
   pvr_pds_build_limm(instr, pvr_pds_ra_high_32b_from_64b(dst), imm_high);
}

struct pvr_pds_reg
pvr_pds_build_convert_32_to_64(const struct pvr_pds_reg reg_low,
                               const struct pvr_pds_reg reg_high)
{
   assert(reg_low.type == PVR_PDS_REG_TYPE_TEMP);
   assert(reg_high.type == PVR_PDS_REG_TYPE_TEMP);
   assert(reg_low.reg_size == PVR_PDS_REG_SIZE_32);
   assert(reg_high.reg_size == PVR_PDS_REG_SIZE_32);
   assert(reg_low.idx + 1 == reg_high.idx);  // Ensure they are consecutive

   struct pvr_pds_reg reg_64 = { .type = PVR_PDS_REG_TYPE_TEMP,
                                 .reg_size = PVR_PDS_REG_SIZE_64,
                                 .idx = reg_low.idx };
   return reg_64;
}

void pvr_pds_build_lgr(ASSERTED const struct pvr_device_info *dev_info,
                       struct pvr_pds_instr *instr,
                       const struct pvr_pds_reg dst,
                       uint32_t src_global_reg_idx)
{
   *instr = (struct pvr_pds_instr){
      .type = PVR_PDS_OP_LIMM,
      .flags = PVR_PDS_INSTR_FLAG_IMM_IS_GR,
   };

   assert(PVR_HAS_FEATURE(dev_info, usc_num_global_registers));
   assert(src_global_reg_idx <
          PVR_GET_FEATURE_VALUE(dev_info, usc_num_global_registers, 0));

   instr->srcs[0] = pvr_pds_instr_src_from_imm(src_global_reg_idx);

   assert(dst.type == PVR_PDS_REG_TYPE_TEMP);
   assert(dst.reg_size == PVR_PDS_REG_SIZE_32);
   instr->srcs[PVR_PDS_INSTR_DST] = pvr_pds_instr_src_from_reg(dst);
}

void pvr_pds_build_ld(struct pvr_pds_instr *instr,
                      const struct pvr_pds_reg src0)
{
   *instr = (struct pvr_pds_instr){
      .type = PVR_PDS_OP_LD,
   };

   assert(src0.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src0.reg_size == pvr_pds_reg_size(LD_SRC0));
   instr->srcs[0] = pvr_pds_instr_src_from_reg(src0);
}

void pvr_pds_build_st(struct pvr_pds_instr *instr,
                      const struct pvr_pds_reg src0)
{
   *instr = (struct pvr_pds_instr){
      .type = PVR_PDS_OP_ST,
   };

   assert(src0.type != PVR_PDS_REG_TYPE_INVALID);
   assert(src0.reg_size == pvr_pds_reg_size(ST_SRC0));
   instr->srcs[0] = pvr_pds_instr_src_from_reg(src0);
}

void pvr_pds_build_wdf(struct pvr_pds_instr *instr)
{
   *instr = (struct pvr_pds_instr){
      .type = PVR_PDS_OP_LIMM,
   };
}

void pvr_pds_build_lock(struct pvr_pds_instr *instr)
{
   *instr = (struct pvr_pds_instr){
      .type = PVR_PDS_OP_LOCK,
   };
}

void pvr_pds_build_release(struct pvr_pds_instr *instr)
{
   *instr = (struct pvr_pds_instr){
      .type = PVR_PDS_OP_RELEASE,
   };
}

void pvr_pds_build_halt(struct pvr_pds_instr *instr)
{
   *instr = (struct pvr_pds_instr){
      .type = PVR_PDS_OP_HALT,
   };
}

void pvr_pds_build_nop(struct pvr_pds_instr *instr)
{
   *instr = (struct pvr_pds_instr){
      .type = PVR_PDS_OP_NOP,
   };
}
