/*
 * Copyright © 2024 Imagination Technologies Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdbool.h>
#include <stdint.h>

#include "pds/pvr_pds_isa.h"
#include "pds/pvr_pds_isa_encode.h"
#include "pvr_device_info.h"
#include "pvr_limits.h"
#include "pvr_pds_builder.h"
#include "pvr_pds_ra.h"
#include "pvr_types.h"
#include "util/u_math.h"
#include "util/macros.h"
#include "vk_alloc.h"
#include "pvr_pds_gen.h"
#include "pvr_private.h"

/* Including this later so that we get the `__pvr_address_type` etc. from
 * pvr_pds_builder.h
 */
#include "csbgen/rogue_pbestate.h"

#define RGX_PDSINST_ST_COUNT4_MAX_SIZE 4
#define PVR_ROGUE_PDSINST_ST_COUNT4_MAX_SIZE PVR_ROGUE_PDSINST_ST_COUNT4_MASK

MUST_CHECK static uint8_t *
pvr_pds_instr_encode_and_advance(struct pvr_pds_instr *instr, uint8_t *buffer)
{
   const size_t bytes_written = pvr_pds_encode_instr(instr, buffer);
   buffer += bytes_written;

   return buffer;
}

void pvr_pds_gen_pixel_event_program(
   enum pvr_pds_builder_generate_mode gen_mode,
   const struct pvr_pds_pixel_event_program_input *input,
   struct pvr_pds_pixel_event_program_output *out)
{
   struct pvr_pds_ra_state ra_state;
   struct pvr_pds_instr instr;

   struct pvr_pds_reg_array doutw_src0_reg_array;
   struct pvr_pds_reg_array doutw_src1_reg_array;
   struct pvr_pds_reg doutu_src0;

   pvr_pds_ra_state_init(&ra_state);

   doutu_src0 = pvr_pds_allocate_src_const(DOUTU_SRC0, &ra_state);

   doutw_src0_reg_array =
      pvr_pds_allocate_src_const_contiguous(DOUTW_SRC0,
                                            &ra_state,
                                            input->emit_word_pair_count);
   doutw_src1_reg_array =
      pvr_pds_allocate_src_const_contiguous(DOUTW_SRC1,
                                            &ra_state,
                                            input->emit_word_pair_count);

   if (gen_mode == PVR_PDS_GENERATE_CODE) {
      uint8_t *buffer = (uint8_t *)out->buffer;

      for (uint32_t i = 0; i < input->emit_word_pair_count; i++) {
         struct pvr_pds_reg doutw_src0 =
            pvr_pds_reg_from_array(doutw_src0_reg_array, i);
         struct pvr_pds_reg doutw_src1 =
            pvr_pds_reg_from_array(doutw_src1_reg_array, i);

         pvr_pds_build_doutw(&instr, doutw_src0, doutw_src1, 0);
         buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);
      }

      pvr_pds_build_doutu(&instr, doutu_src0, PVR_PDS_INSTR_FLAG_END);
      buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);
   } else if (gen_mode == PVR_PDS_GENERATE_DATA) {
      const uint32_t emit_word_size =
         ROGUE_PBESTATE_STATE_WORD0_length + ROGUE_PBESTATE_STATE_WORD1_length;
      uint32_t *buffer = (uint32_t *)out->buffer;

      pvr_pds_src_write(DOUTU_SRC0,
                        buffer,
                        doutu_src0,
                        &input->data.task_control);

      pvr_pds_src_write_multiple(DOUTW_SRC0,
                                 buffer,
                                 &doutw_src0_reg_array,
                                 input->data.emit_words,
                                 input->emit_word_pair_count);

      for (uint32_t i = 0; i < input->emit_word_pair_count; i++) {
         const struct pvr_pds_reg doutw_src1 =
            pvr_pds_reg_from_array(doutw_src1_reg_array, i);
         const bool is_last = (i == input->emit_word_pair_count - 1);
         uint32_t doutw_src1_value;

         /* The placement in the USC common store relies on an implicit layout.
          * We're expecting the EOT shader to also utilize this layout.
          */
         pvr_instr_src_pack (&doutw_src1_value, DOUTW_SRC1, val) {
            val.a0 = emit_word_size * i;
            val.bsize = pvr_pds_doutw_src1_bsize_from_uint(emit_word_size);
            val.dest = ROGUE_PDS_INSTR_SRC_DOUTD_DEST_COMMON_STORE;

            if (PVR_HAS_FEATURE(input->data.dev_info, slc_mcu_cache_controls))
               val.cmode = ROGUE_PDS_INSTR_SRC_CMODE_LD_BYPASS;

            val.last = is_last;
         }

         pvr_pds_src_write(DOUTW_SRC1, buffer, doutw_src1, &doutw_src1_value);
      }
   }

   out->data_size = pvr_pds_ra_get_allocated_32b_consts(&ra_state);

   /* 1 DOUTW per emit pair, and final DOUTU to kick the EOT. */
   out->code_size = input->emit_word_pair_count + 1;
}

struct pvr_pds_ld_memory_region_setup_info {
   struct pvr_pds_reg_array const_reg_array;
   /* COUNT8 value for each LD apart the last one. */
   uint32_t ld_count8;
   /* COUNT8 value for the last LD. */
   uint32_t last_ld_count8;
};

struct pvr_pds_st_memory_region_setup_info {
   struct pvr_pds_reg_array const_reg_array;
   /* COUNT4 value for each ST apart the last one. */
   uint32_t st_count4;
   /* COUNT4 value for the last ST. */
   uint32_t last_st_count4;
};

static void pvr_pds_ld_memory_region_setup(
   struct pvr_pds_ra_state *ra_state,
   const struct pvr_pds_reg_array dst_reg_array,
   struct pvr_pds_ld_memory_region_setup_info *info_out)
{
   // FIXME: This is the buffer size, not the register count?
   const uint32_t size_in_64b_reg = dst_reg_array.reg_count;
   struct pvr_pds_reg_array const_reg_array;
   uint32_t ld_instr_count;
   uint32_t last_ld_count8;
   uint32_t ld_count8;

   /* If this is hit, we might want to add support for loading into 32b regs. */
   assert(dst_reg_array.reg0.reg_size == PVR_PDS_REG_SIZE_64);

   if (size_in_64b_reg <= 7) {
      ld_instr_count = 1;
      last_ld_count8 = size_in_64b_reg;
      ld_count8 = size_in_64b_reg;
   } else {
#define LD_SRC_ADDR_U64_ALIGNMENT 2

      const uint32_t ld_count8_max = pvr_instr_src(LD_SRC0_COUNT8_MAX);
      /* We might not be able to use COUNT8_MAX in case it results in subsequent
       * LD having unaligned source addresses.
       */
      const uint32_t size_in_u64_per_ld =
         ROUND_DOWN_TO(ld_count8_max, LD_SRC_ADDR_U64_ALIGNMENT);
      uint32_t remaining_u64_count;

      static_assert(LD_SRC_ADDR_U64_ALIGNMENT ==
                       pvr_instr_src(LD_SRC0_SRCADD_ALIGNMENT) / 8,
                    "#define and the code below needs updating.");

      ld_count8 = size_in_u64_per_ld;

      ld_instr_count = size_in_64b_reg / size_in_u64_per_ld;
      remaining_u64_count =
         size_in_64b_reg - ld_instr_count * size_in_u64_per_ld;

      if (remaining_u64_count == 0) {
         last_ld_count8 = size_in_u64_per_ld;
      } else if (remaining_u64_count == 1 &&
                 ld_count8_max % LD_SRC_ADDR_U64_ALIGNMENT == 1) {
         /* We can squeeze this into the previous LD since we don't need to
          * worry about following LDs containing unaligned source addresses
          * since this will be the last one.
          */
         last_ld_count8 = 7;
      } else {
         ld_instr_count += 1;
         last_ld_count8 = remaining_u64_count;
      }

#undef LD_SRC_ADDR_U64_ALIGNMENT
   }

   const_reg_array =
      pvr_pds_allocate_src_const_contiguous(LD_SRC0, ra_state, ld_instr_count);

   *info_out = (struct pvr_pds_ld_memory_region_setup_info){
      .const_reg_array = const_reg_array,
      .ld_count8 = ld_count8,
      .last_ld_count8 = last_ld_count8,
   };
}

static void pvr_pds_st_memory_region_setup(
   struct pvr_pds_ra_state *ra_state,
   const struct pvr_pds_reg_array dst_reg_array,
   struct pvr_pds_st_memory_region_setup_info *info_out)
{
   // FIXME: check the register or memory block size
   const uint32_t size_in_32b_reg = dst_reg_array.reg_count;
   struct pvr_pds_reg_array const_reg_array;
   uint32_t st_instr_count;
   uint32_t last_st_count4;
   uint32_t st_count4;

   /* Calculate number of ST instructions needed */
   st_instr_count = size_in_32b_reg / RGX_PDSINST_ST_COUNT4_MAX_SIZE;
   st_instr_count +=
      ((size_in_32b_reg % RGX_PDSINST_ST_COUNT4_MAX_SIZE) == 0 ? 0 : 1);

   /* Initialize constants */
   const_reg_array =
      pvr_pds_allocate_src_const_contiguous(ST_SRC0, ra_state, st_instr_count);

   // FIXME: Make sure these calculations come before the register allocation
   /* Calculate ST counts */
   st_count4 = (size_in_32b_reg <= RGX_PDSINST_ST_COUNT4_MAX_SIZE
                   ? size_in_32b_reg
                   : RGX_PDSINST_ST_COUNT4_MAX_SIZE);

   /* Calculate last ST count */
   last_st_count4 = size_in_32b_reg - st_count4 * (st_instr_count - 1);

   /* Fill info_out struct */
   *info_out = (struct pvr_pds_st_memory_region_setup_info){
      .const_reg_array = const_reg_array,
      .st_count4 = st_count4,
      .last_st_count4 = last_st_count4,
   };
}

static uint8_t *pvr_pds_ld_memory_region_code(
   const struct pvr_pds_ld_memory_region_setup_info *info,
   uint8_t *buffer)
{
   /* Each LD has a single source so the amount of consts is also the amount of
    * LD instructions.
    */
   const uint32_t ld_instr_count = info->const_reg_array.reg_count;
   struct pvr_pds_instr instr;

   for (uint32_t i = 0; i < ld_instr_count; i++) {
      const struct pvr_pds_reg src0_reg =
         pvr_pds_reg_from_array(info->const_reg_array, i);

      pvr_pds_build_ld(&instr, src0_reg);
      buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);
   }

   return buffer;
}

static uint8_t *pvr_pds_st_memory_region_code(
   const struct pvr_pds_st_memory_region_setup_info *info,
   uint8_t *buffer)
{
   const uint32_t st_instr_count = info->const_reg_array.reg_count;
   struct pvr_pds_instr instr;

   for (uint32_t i = 0; i < st_instr_count; i++) {
      const struct pvr_pds_reg src0_reg =
         pvr_pds_reg_from_array(info->const_reg_array, i);
      pvr_pds_build_st(&instr, src0_reg);

      buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);
   }

   return buffer;
}

static void pvr_pds_ld_memory_region_data(
   const struct pvr_device_info *dev_info,
   const struct pvr_pds_ld_memory_region_setup_info *info,
   pvr_dev_addr_t src_address,
   const struct pvr_pds_reg_array dest_array,
   uint32_t *buffer)
{
   /* Each LD has a single source so the amount of consts is also the amount of
    * LD instructions.
    */
   const uint32_t ld_instr_count = info->const_reg_array.reg_count;

   for (uint32_t i = 0; i < ld_instr_count; i++) {
      const struct pvr_pds_reg ld_src0 =
         pvr_pds_reg_from_array(info->const_reg_array, i);
      uint64_t const_val;

      pvr_instr_src_pack (&const_val, LD_SRC0, val) {
         val.cmode = pvr_instr_src(CMODE_LD_CACHED);
         if (PVR_HAS_FEATURE(dev_info, slc_mcu_cache_controls))
            val.slcmode = pvr_instr_src(SLC_MODE_LD_CACHED);

         val.dest = pvr_pds_reg_from_array(dest_array, i * info->ld_count8).idx;

         if (i == ld_instr_count - 1)
            val.count8 = info->last_ld_count8;
         else
            val.count8 = info->ld_count8;

         val.srcadd = src_address;
         src_address = PVR_DEV_ADDR_OFFSET(src_address, val.count8 * 8);
      }

      pvr_pds_src_write(LD_SRC0, buffer, ld_src0, &const_val);
   }
}

static void pvr_pds_st_memory_region_data(
   const struct pvr_device_info *dev_info,
   const struct pvr_pds_st_memory_region_setup_info *info,
   pvr_dev_addr_t dst_address,
   const struct pvr_pds_reg_array dest_array,
   uint32_t *buffer)
{
   const uint32_t st_instr_count = info->const_reg_array.reg_count;

   for (uint32_t i = 0; i < st_instr_count; i++) {
      const struct pvr_pds_reg st_src0 =
         pvr_pds_reg_from_array(info->const_reg_array, i);
      uint64_t const_val;

      pvr_instr_src_pack (&const_val, ST_SRC0, val) {
         if (PVR_HAS_FEATURE(dev_info, slc_mcu_cache_controls))
            val.slcmode = pvr_instr_src(SLC_MODE_LD_CACHED);
         val.src = pvr_pds_reg_from_array(dest_array, i * info->st_count4).idx;
         val.cmode = pvr_instr_src(CMODE_ST_WRITE_THROUGH);
         val.count4 =
            (info->last_st_count4 / PVR_ROGUE_PDSINST_ST_COUNT4_MAX_SIZE);
         val.count4 +=
            (info->last_st_count4 % PVR_ROGUE_PDSINST_ST_COUNT4_MAX_SIZE) == 0
               ? 0
               : 1;
         val.dstadd = dst_address;
         dst_address = PVR_DEV_ADDR_OFFSET(dst_address, val.count4 * 8);
      }

      pvr_pds_src_write(ST_SRC0, buffer, st_src0, &const_val);
   }
}

void pvr_pds_gen_xfb_program(enum pvr_pds_builder_generate_mode gen_mode,
                             const struct pvr_pds_xfb_program_input *input,
                             struct pvr_pds_xfb_program_output *out)
{
   struct pvr_pds_reg_array so_const_reg_array;
   struct pvr_pds_reg_array so_temp_reg_array;
   struct pvr_pds_ra_state ra_state;
   struct pvr_pds_instr instr;

   /* TODO: Support needs to be added here.
    *
    * With this feature we have access to `STMP` so use that instead of
    * `STM`.
    */
   assert(!PVR_HAS_FEATURE(input->dev_info, parallel_stream_out));

   /* TODO: Support needs to be added here.
    *
    * SO_GLOBAL is triggered on any stream out overflow and the PDS will
    * abort all stream out operations, so for parallel stream out it might
    * halt processing of an earlier stream out.
    *
    * So instead of using SO_GLOBAL, we'll check if the stream out will
    * overflow and set P0 accordingly.
    */
   assert(!PVR_HAS_FEATURE(input->dev_info, parallel_stream_out) ||
          !PVR_HAS_QUIRK(input->dev_info, 50853));

   pvr_pds_ra_state_init(&ra_state);

   static_assert(pvr_instr_src_length(SO_SRC0) == 2, "");
   static_assert(pvr_instr_src_length(SO_SRC1) == 2, "");
   static_assert(pvr_instr_src_length(SO_SRC3) == 2, "");

   // FIXME: Make sure this is not conflicting with other ptemp registers
   so_temp_reg_array = pvr_pds_ra_allocate_64b_persistent_temp_contiguous(
      &ra_state,
      3 * input->varying_count);

   so_const_reg_array =
      pvr_pds_allocate_src_const_contiguous(SO_SRC2,
                                            &ra_state,
                                            input->varying_count);

   if (gen_mode == PVR_PDS_GENERATE_CODE) {
      uint8_t *buffer = (uint8_t *)out->buffer;

      if (input->test_overflow) {
         for (uint32_t i = 0; i < input->varying_count; i++) {
            const struct pvr_pds_reg so_src0_reg =
               pvr_pds_reg_from_array(so_temp_reg_array, i * 3);
            const struct pvr_pds_reg so_src1_reg =
               pvr_pds_reg_from_array(so_temp_reg_array, i * 3 + 1);
            const struct pvr_pds_reg so_src3_reg =
               pvr_pds_reg_from_array(so_temp_reg_array, i * 3 + 2);

            const struct pvr_pds_reg so_src2_reg =
               pvr_pds_reg_from_array(so_const_reg_array, i);

            static_assert(PVR_MAX_XFB_STREAMS == 1,
                          "This program likely needs changing");

            pvr_pds_build_stm(&instr,
                              0, /* Always stream out to stream 0. */
                              so_src0_reg,
                              so_src1_reg,
                              so_src2_reg,
                              so_src3_reg,
                              PVR_PDS_INSTR_FLAG_TST);
            instr.pred.so_global = i > 0;

            buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);
         }
      }

      for (uint32_t i = 0; i < input->varying_count; i++) {
         const struct pvr_pds_reg so_src0_reg =
            pvr_pds_reg_from_array(so_temp_reg_array, i * 3);
         const struct pvr_pds_reg so_src1_reg =
            pvr_pds_reg_from_array(so_temp_reg_array, i * 3 + 1);
         const struct pvr_pds_reg so_src3_reg =
            pvr_pds_reg_from_array(so_temp_reg_array, i * 3 + 2);

         const struct pvr_pds_reg so_src2_reg =
            pvr_pds_reg_from_array(so_const_reg_array, i);

         static_assert(PVR_MAX_XFB_STREAMS == 1,
                       "This program likely needs changing");

         pvr_pds_build_stm(&instr,
                           0, /* Always stream out to stream 0. */
                           so_src0_reg,
                           so_src1_reg,
                           so_src2_reg,
                           so_src3_reg,
                           0);

         /* No need to clear this with STMC since it should have been done by
          * the transform feedback initialization program beforehand.
          */
         instr.pred.so_global = true;

         buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);
      }
      pvr_pds_build_halt(&instr);
      buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);
   } else if (gen_mode == PVR_PDS_GENERATE_DATA) {
      uint32_t *buffer = (uint32_t *)out->buffer;

      for (uint32_t i = 0; i < input->varying_count; i++) {
         const struct pvr_pds_xfb_varying_info *varying_info =
            &input->data.varying_info_arr[i];
         const struct pvr_pds_reg so_src2_reg =
            pvr_pds_reg_from_array(so_const_reg_array, i);
         uint32_t so_src2_val;

         pvr_instr_src_pack (&so_src2_val, SO_SRC2, val) {
            val.so_vioff = varying_info->input_offset;
            val.so_dmasize = 0;
            val.so_vooff = varying_info->output_offset;
         }

         pvr_pds_src_write(SO_SRC2, buffer, so_src2_reg, &so_src2_val);
      }
   }

   out->data_size = pvr_pds_ra_get_allocated_32b_consts(&ra_state);

   /* STM with optional STM(TST) for overflow testing. */
   out->code_size = input->varying_count * (input->test_overflow ? 2 : 1);
   out->code_size += 1; /* HALT, terminating the program. */

   out->temp_size = 3 * input->varying_count;
}

/* Init value of persistent temps which WILL be used in STM *
 * instructions in transform feedback PDS program.          */
static void
pvr_pds_xfb_ptem_buffer_init(const struct pvr_pds_xfb_init_program_input *input,
                             const struct pvr_pds_reg_array *ptemp_array)
{
   const struct pvr_device_info *dev_info = input->dev_info;
   pvr_dev_addr_t buffer = input->data.ptemp_memory_buffer_addr; // FIXME: Cast this buffer
   uint16_t primitive_type = input->primitive_type;

   /* Updating the persistent temp buffer */
   for (uint32_t i = 0; i < ptemp_array->reg_count / 3; i++) {
      pvr_instr_src_pack (&buffer.addr, SO_SRC0, src0) {
         const struct pvr_pds_reg so_src0_reg =
            pvr_pds_reg_from_array(*ptemp_array, i * 3);
         const struct pvr_pds_reg so_src1_reg =
            pvr_pds_reg_from_array(*ptemp_array, i * 3 + 1);

         if (PVR_HAS_FEATURE(dev_info, slc_mcu_cache_controls))
            src0.slcmode = pvr_instr_src(SLC_MODE_ST_WRITE_BACK);

         /* Note: SO_SRC3 follows SO_SRC1 in the ptems that's why the `* 3`. The
          * hw requires SO_SRC3 to be either in temps or ptems, so we just used
          * ptemps there.
          */

         /* SO_SRC{0,1} are loaded into persistent temps themselves so the ptemp
          * index matches the index of so_src0 rather than being a separate
          * ptemp register allocation.
          */
         src0.ptemp_so_prim = so_src1_reg.idx;
         src0.ptemp_so_address = so_src0_reg.idx;

         src0.so_address = input->data.buffer_info[i].buffer_address;
      }
      buffer.addr += pvr_instr_src_length(SO_SRC0);

      pvr_instr_src_pack (&buffer.addr, SO_SRC1, src1) {
         src1.so_primneeded = 0;
         src1.so_primwritten = 0;
      }
      buffer.addr += pvr_instr_src_length(SO_SRC1);

      pvr_instr_src_pack (&buffer.addr, SO_SRC3, src3) {
         uint32_t vo_size = input->data.varying_info_arr[i].varying_size;
         src3.so_primtype = primitive_type;
         src3.so_vosize = vo_size;

         // TODO: Support multiple STMs for vertex data that can't be streamed
         // out with just a single STM. We'll need to setup another SO_SRC3 with
         // so_eop set to false and used that in the STMs until the last one
         // which should have eop set.
         src3.so_eop = true;

         src3.so_limit = input->data.buffer_info[i].buffer_size;
      }
      buffer.addr += pvr_instr_src_length(SO_SRC3);
   }
}

void pvr_pds_gen_xfb_init_program(
   enum pvr_pds_builder_generate_mode gen_mode,
   const struct pvr_pds_xfb_init_program_input *input,
   struct pvr_pds_xfb_program_output *out)
{
   uint32_t counter_buffers_count;
   struct pvr_pds_ra_state ra_state;
   struct pvr_pds_reg scratch_reg, counter_buff_reg;
   struct pvr_pds_reg_array scratch_reg_arr, counter_buffer_add_reg,
      counter_buff_offset_reg_arr;
   struct pvr_pds_ld_memory_region_setup_info ptemp_ld_setup_info;
   struct pvr_pds_ld_memory_region_setup_info counter_buff_addr_ld_setup_info;
   struct pvr_pds_st_memory_region_setup_info counter_pair_st_setup_info;
   struct pvr_pds_st_memory_region_setup_info counter_buff_offset_st_setup_info;

   pvr_pds_ra_state_init(&ra_state);

   counter_buffers_count = input->counter_buffer_count;

   struct pvr_pds_reg temp64_r1 = pvr_pds_ra_allocate_64b_temp(&ra_state);
   struct pvr_pds_reg temp64_r2 = pvr_pds_ra_allocate_64b_temp(&ra_state);
   struct pvr_pds_reg temp32_r1 = pvr_pds_ra_allocate_32b_temp(&ra_state);
   struct pvr_pds_reg temp32_r2 = pvr_pds_ra_allocate_32b_temp(&ra_state);

   // Allocate necessary registers
   scratch_reg_arr =
      pvr_pds_ra_allocate_64b_temp_contiguous(&ra_state, PVR_MAX_XFB_BUFFERS);
   counter_buff_offset_reg_arr =
      pvr_pds_ra_allocate_64b_temp_contiguous(&ra_state, 1);
   counter_buffer_add_reg =
      pvr_pds_ra_allocate_64b_temp_contiguous(&ra_state, counter_buffers_count);

   struct pvr_pds_reg_array ptemp_reg_array =
      pvr_pds_ra_allocate_64b_persistent_temp_contiguous(
         &ra_state,
         input->persistent_temp_size);

   // Setup memory regions
   pvr_pds_ld_memory_region_setup(&ra_state,
                                  ptemp_reg_array,
                                  &ptemp_ld_setup_info);
   pvr_pds_ld_memory_region_setup(&ra_state,
                                  counter_buffer_add_reg,
                                  &counter_buff_addr_ld_setup_info);
   pvr_pds_st_memory_region_setup(&ra_state,
                                  scratch_reg_arr,
                                  &counter_pair_st_setup_info);
   pvr_pds_st_memory_region_setup(&ra_state,
                                  counter_buff_offset_reg_arr,
                                  &counter_buff_offset_st_setup_info);

   if (gen_mode == PVR_PDS_GENERATE_CODE) {
      uint8_t *buffer = (uint8_t *)out->buffer;
      struct pvr_pds_instr instr;

      pvr_pds_build_stmc(&instr, PVR_PDS_SO_MASK_GLOBAL);
      buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);

      // Load varying counter addresses into persistent temps
      pvr_pds_ld_memory_region_code(&ptemp_ld_setup_info, buffer);

      for (uint32_t i = 0; i < PVR_MAX_XFB_BUFFERS; ++i) {
         scratch_reg = pvr_pds_reg_from_array(scratch_reg_arr, i);
         buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);
      }

      buffer =
         pvr_pds_st_memory_region_code(&counter_pair_st_setup_info, buffer);

      if (input->load_from_counter_buffers) {
         pvr_pds_ld_memory_region_code(&counter_buff_addr_ld_setup_info,
                                       buffer);

         for (uint32_t i = 0; i < counter_buffers_count; i++) {
            scratch_reg = pvr_pds_reg_from_array(scratch_reg_arr, i);
            counter_buff_reg =
               pvr_pds_reg_from_array(counter_buffer_add_reg, i);
            pvr_pds_build_mov64(&instr, scratch_reg, counter_buff_reg);
         }

         // Add counters (byte offsets) to the corresponding addresses in ptemps
         for (uint32_t i = 0; i < input->num_loads; ++i) {
            struct pvr_pds_reg ptemp_reg =
               pvr_pds_reg_from_array(ptemp_reg_array, i);

            // Counters are in bytes, so shift down to dwords
            pvr_pds_build_limm(&instr, temp32_r1, input->load_dest_select[i]);
            buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);

            pvr_pds_build_add32(
               input->dev_info,
               &instr,
               temp32_r1,
               temp32_r1,
               pvr_pds_ra_low_32b_from_64b(pvr_pds_reg_from_array(
                  counter_pair_st_setup_info.const_reg_array,
                  input->load_dest_select[i])),
               PVR_PDS_INSTR_FLAG_SIGNED_OP);
            buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);

            pvr_pds_build_limm(&instr, temp32_r2, 0x1);
            buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);

            pvr_pds_build_mov64(&instr, temp64_r2, ptemp_reg);
            buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);

            pvr_pds_build_mad(&instr,
                              temp64_r1,
                              temp32_r1,
                              temp32_r2,
                              ptemp_reg,
                              PVR_PDS_INSTR_FLAG_SIGNED_OP);
            buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);

            pvr_pds_build_mov64(&instr, ptemp_reg, temp64_r1);
            buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);
         }
      }

      buffer = pvr_pds_st_memory_region_code(&counter_buff_offset_st_setup_info,
                                             buffer);
   } else if (gen_mode == PVR_PDS_GENERATE_DATA) {
      uint32_t *buffer = (uint32_t *)out->buffer;

      pvr_pds_xfb_ptem_buffer_init(input, &ptemp_reg_array);

      pvr_pds_ld_memory_region_data(input->dev_info,
                                    &ptemp_ld_setup_info,
                                    input->data.ptemp_memory_buffer_addr,
                                    ptemp_reg_array,
                                    buffer);
      pvr_pds_st_memory_region_data(input->dev_info,
                                    &counter_pair_st_setup_info,
                                    input->data.counter_pairs_address[0],
                                    scratch_reg_arr,
                                    buffer);

      if (input->load_from_counter_buffers) {
         pvr_pds_ld_memory_region_data(input->dev_info,
                                       &counter_buff_addr_ld_setup_info,
                                       input->data.counter_buffers_addr[0],
                                       counter_buffer_add_reg,
                                       buffer);
         pvr_pds_st_memory_region_data(
            input->dev_info,
            &counter_buff_offset_st_setup_info,
            input->data.counter_buffer_offsets_addr[0],
            counter_buff_offset_reg_arr,
            buffer);
      }
   }

   out->data_size = pvr_pds_ra_get_allocated_32b_consts(&ra_state);
   out->code_size = 1 + PVR_MAX_XFB_BUFFERS + ptemp_ld_setup_info.ld_count8 +
                     counter_pair_st_setup_info.st_count4 +
                     counter_buff_addr_ld_setup_info.ld_count8 +
                     counter_buff_offset_st_setup_info.st_count4;

   if (input->load_from_counter_buffers) {
      out->code_size += counter_buffers_count + input->num_loads * 5 +
                        counter_buff_addr_ld_setup_info.ld_count8;
   }

   out->temp_size = scratch_reg_arr.reg_count +
                    counter_buffer_add_reg.reg_count +
                    counter_buff_offset_reg_arr.reg_count;
}

void pvr_pds_gen_transform_end_program(
   enum pvr_pds_builder_generate_mode gen_mode,
   const struct pvr_pds_xfb_end_program_input *input,
   struct pvr_pds_xfb_program_output *out)
{
   struct pvr_pds_ra_state ra_state;
   pvr_pds_ra_state_init(&ra_state);

   struct pvr_pds_reg scratch_temp, written_count_temp;
   struct pvr_pds_reg buffer_stride_reg;
   struct pvr_pds_reg_array counter_buff_offset_reg_arr,
      counter_buff_addr_reg_arr, count_pair_reg_arr;

   struct pvr_pds_ld_memory_region_setup_info counter_pair_ld_setup_info,
      counter_buff_offset_setup_info;
   struct pvr_pds_st_memory_region_setup_info counter_buff_addr_setup_info;

   scratch_temp = pvr_pds_ra_allocate_64b_temp(&ra_state);
   written_count_temp = pvr_pds_ra_allocate_64b_temp(&ra_state);
   buffer_stride_reg = pvr_pds_ra_allocate_64b_temp(&ra_state);

   count_pair_reg_arr = pvr_pds_ra_allocate_64b_temp_contiguous(&ra_state, 1);
   counter_buff_offset_reg_arr =
      pvr_pds_ra_allocate_64b_temp_contiguous(&ra_state, 1);
   counter_buff_addr_reg_arr =
      pvr_pds_ra_allocate_64b_temp_contiguous(&ra_state, 1);

   pvr_pds_ld_memory_region_setup(&ra_state,
                                  count_pair_reg_arr,
                                  &counter_pair_ld_setup_info);

   pvr_pds_ld_memory_region_setup(&ra_state,
                                  counter_buff_offset_reg_arr,
                                  &counter_buff_offset_setup_info);

   pvr_pds_st_memory_region_setup(&ra_state,
                                  counter_buff_addr_reg_arr,
                                  &counter_buff_addr_setup_info);

   if (gen_mode == PVR_PDS_GENERATE_CODE) {
      uint8_t *buffer = (uint8_t *)out->buffer;
      struct pvr_pds_instr instr;

      if (input->store_counter_buffers)
         pvr_pds_build_limm64(&instr, written_count_temp, 1U);

      if (input->store_counter_buffers) {
         /* LD counter_pairs_address */
         pvr_pds_ld_memory_region_code(&counter_pair_ld_setup_info, buffer);
         pvr_pds_build_add64(&instr,
                             written_count_temp,
                             pvr_pds_reg_from_array(count_pair_reg_arr, 0),
                             written_count_temp,
                             PVR_PDS_INSTR_FLAG_INVERT);

         pvr_pds_ld_memory_region_code(&counter_buff_offset_setup_info, buffer);

         for (uint8_t i = 0; i < input->counter_buffer_count; i++) {
            /* If the API has a counter buffer which needs to be updated */
            if (input->data.counter_buffers_addr[i].addr) {
               /* Convert the counter offset to a 64-bit value */
               pvr_pds_build_mov64(
                  &instr,
                  scratch_temp,
                  pvr_pds_reg_from_array(counter_buff_offset_reg_arr, 0));
               buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);

               if (input->buffer_strides[i]) {
                  pvr_pds_build_limm64(&instr,
                                     buffer_stride_reg,
                                     input->buffer_strides[i]);
                  pvr_pds_build_mad(&instr,
                                    scratch_temp,
                                    written_count_temp,
                                    buffer_stride_reg,
                                    scratch_temp,
                                    PVR_PDS_INSTR_FLAG_SIGNED_OP);
                  buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);
               }
               /* ST counter_buffers_addr*/
               buffer =
                  pvr_pds_st_memory_region_code(&counter_buff_addr_setup_info,
                                                buffer);
            }
         }
      }

   } else if (gen_mode == PVR_PDS_GENERATE_DATA) {
      uint32_t *buffer = (uint32_t *)out->buffer;
      /* Setup LD_SRC for counter_pair_addr */
      pvr_pds_ld_memory_region_data(input->dev_info,
                                    &counter_pair_ld_setup_info,
                                    input->data.counter_pairs_address[0],
                                    count_pair_reg_arr,
                                    buffer);

      /* LD counter_buffer_offset_addr */
      pvr_pds_ld_memory_region_data(input->dev_info,
                                    &counter_buff_offset_setup_info,
                                    input->data.counter_buffer_offsets_addr[0],
                                    counter_buff_offset_reg_arr,
                                    buffer);

      /* ST in counter_buffer_addr */
      for (uint8_t i = 0; i < input->counter_buffer_count; i++) {
         pvr_pds_st_memory_region_data(
            input->dev_info,
            &counter_buff_addr_setup_info,
            input->data.counter_buffer_offsets_addr[0],
            counter_buff_addr_reg_arr,
            buffer);
      }
   }

   /* data segment size */
   out->data_size = pvr_pds_ra_get_allocated_32b_consts(&ra_state);

   /* Calculate the code_size */
   if (input->store_counter_buffers) {
      out->code_size = 3;
      out->code_size += 1;
      out->code_size += input->counter_buffer_count;
      out->code_size += counter_pair_ld_setup_info.ld_count8;
      out->code_size += counter_buff_offset_setup_info.ld_count8;
      for (uint32_t i = 0; i < input->counter_buffer_count; i++) {
         if (input->data.counter_buffers_addr[i].addr) {
            out->code_size += 1;
            out->code_size += counter_buff_addr_setup_info.st_count4;
            if (input->buffer_strides[i]) {
               out->code_size += 2;
            }
         }
      }
   }

   /* temp registers */
   out->temp_size = 3;
   out->temp_size += count_pair_reg_arr.reg_count;
   out->temp_size += counter_buff_offset_reg_arr.reg_count;
   out->temp_size += counter_buff_addr_reg_arr.reg_count;
}

void pvr_pds_gen_generate_vertex_count_program(
   enum pvr_pds_builder_generate_mode gen_mode,
   const struct pvr_pds_xfb_draw_program_input *input,
   struct pvr_pds_xfb_program_output *out)
{
   struct pvr_pds_ld_memory_region_setup_info
      last_captured_primitive_ld_setup_info = { 0 };
   struct pvr_pds_st_memory_region_setup_info
      out_counter_buffer_st_setup_info = { 0 };
   struct pvr_pds_ra_state ra_state;
   struct pvr_pds_reg temp_reg, counter_buffer_addr, neg_counter_offset_reg;
   struct pvr_pds_reg_array temp_reg_array, out_count_buffer;
   uint8_t index = 0;

   pvr_pds_ra_state_init(&ra_state);

   /* Array of temps - vertex count, Instance count, first vertex, first *
    * instance                                                           */
   temp_reg_array = pvr_pds_ra_allocate_32b_temp_contiguous(&ra_state, 4);
   out_count_buffer = pvr_pds_ra_allocate_32b_temp_contiguous(&ra_state, 1);
   neg_counter_offset_reg = pvr_pds_ra_allocate_32b_temp(&ra_state);

   pvr_pds_ld_memory_region_setup(&ra_state,
                                  temp_reg_array,
                                  &last_captured_primitive_ld_setup_info);

   pvr_pds_st_memory_region_setup(&ra_state,
                                  out_count_buffer,
                                  &out_counter_buffer_st_setup_info);

   if (gen_mode == PVR_PDS_GENERATE_CODE) {
      uint8_t *buffer = (uint8_t *)out->buffer;
      struct pvr_pds_instr instr;

      /* LD counter_buffer_addr */
      pvr_pds_ld_memory_region_code(&last_captured_primitive_ld_setup_info,
                                    buffer);

      if (input->negative_counter_offset) {
         temp_reg = pvr_pds_reg_from_array(temp_reg_array, index);
         pvr_pds_build_limm(&instr,
                            neg_counter_offset_reg,
                            input->negative_counter_offset);
         pvr_pds_build_add64(&instr,
                             temp_reg,
                             temp_reg,
                             neg_counter_offset_reg,
                             PVR_PDS_INSTR_FLAG_INVERT);
         buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);
      }

      if (input->vertex_stride) {
         /* IDIV operation */
         struct pvr_pds_reg_array temps =
            pvr_pds_ra_allocate_32b_temp_contiguous(&ra_state, 3);

         uint32_t divisor = input->vertex_stride;

         /* Find Shift Amount */
         uint32_t shift = __builtin_clz(divisor - 1);
         pvr_pds_build_limm(&instr, pvr_pds_reg_from_array(temps, 0), shift);

         /* Calculate Right-Shift Amount for Multiply */
         pvr_pds_build_limm(&instr, pvr_pds_reg_from_array(temps, 1), 31);
         pvr_pds_build_add32(input->dev_info,
                             &instr,
                             pvr_pds_reg_from_array(temps, 0),
                             pvr_pds_reg_from_array(temps, 0),
                             pvr_pds_reg_from_array(temps, 1),
                             PVR_PDS_INSTR_FLAG_SIGNED_OP);

         /* Calculate Multiplier */
         uint64_t multiplier64 =
            ((uint64_t)1 << (shift + 31)) + (divisor - 1) / divisor;
         uint32_t multiplier = (uint32_t)multiplier64;

         pvr_pds_build_limm(&instr,
                            pvr_pds_reg_from_array(temps, 2),
                            multiplier);

         /* Perform Multiply and Shift */
         pvr_pds_build_mad(&instr,
                           pvr_pds_reg_from_array(temps, 2),
                           pvr_pds_reg_from_array(temp_reg_array, index),
                           pvr_pds_reg_from_array(temps, 2),
                           pvr_pds_reg_from_array(temp_reg_array, index),
                           PVR_PDS_INSTR_FLAG_SIGNED_OP);
         pvr_pds_build_sftlp64i(&instr,
                                PVR_PDS_LOGIC_OP_NONE,
                                pvr_pds_reg_from_array(temps, 2),
                                pvr_pds_reg_from_array(temps, 2),
                                pvr_pds_reg_from_array(temps, 0),
                                shift);

         /* Store the Result */
         pvr_pds_build_mov64(&instr,
                             pvr_pds_reg_from_array(temp_reg_array, index),
                             pvr_pds_reg_from_array(temps, 2));
      }
      index++;

      temp_reg = pvr_pds_reg_from_array(temp_reg_array, index);
      index++;
      pvr_pds_build_limm(&instr, temp_reg, input->instance_count);
      buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);

      temp_reg = pvr_pds_reg_from_array(temp_reg_array, index);
      index++;
      pvr_pds_build_limm(&instr, temp_reg, 0);
      buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);

      temp_reg = pvr_pds_reg_from_array(temp_reg_array, index);
      index++;
      pvr_pds_build_limm(&instr, temp_reg, input->first_instance);
      buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);

      /* ST ouput_count_buffer_addr */
      pvr_pds_st_memory_region_code(&out_counter_buffer_st_setup_info, buffer);

      pvr_pds_build_halt(&instr);
   } else if (gen_mode == PVR_PDS_GENERATE_DATA) {
      uint32_t *buffer = (uint32_t *)out->buffer;

      pvr_pds_ld_memory_region_data(input->dev_info,
                                    &last_captured_primitive_ld_setup_info,
                                    input->data.counter_buffer_addr,
                                    temp_reg_array,
                                    buffer);

      pvr_pds_st_memory_region_data(input->dev_info,
                                    &out_counter_buffer_st_setup_info,
                                    input->data.ouput_count_buffer_addr,
                                    out_count_buffer,
                                    buffer);

      out->buffer = buffer;
   }

   /* Calculate the sizes */
   out->data_size = pvr_pds_ra_get_allocated_32b_consts(&ra_state);

   out->code_size = last_captured_primitive_ld_setup_info.ld_count8;
   if (input->negative_counter_offset) {
      out->code_size += 2;
   }
   out->code_size += 4;
   out->code_size += out_counter_buffer_st_setup_info.st_count4;

   out->temp_size = temp_reg_array.reg_count;
   out->temp_size += out_count_buffer.reg_count;
   out->temp_size += 1;
   out->temp_size += 3; /* temps for IDIV */
}

void pvr_pds_gen_generate_query_program(
   enum pvr_pds_builder_generate_mode gen_mode,
   const struct pvr_pds_xfb_query_program_input *input,
   struct pvr_pds_xfb_program_output *out)
{
   struct pvr_pds_st_memory_region_setup_info offset_count_addr_st_setup_info;
   struct pvr_pds_ld_memory_region_setup_info total_count_addr_ld_setup_info;
   struct pvr_pds_ld_memory_region_setup_info ptemp_ld_setup_info;
   struct pvr_pds_reg_array offset_reg_arr;
   struct pvr_pds_reg_array total_reg_arr;
   struct pvr_pds_reg_array ptemp_reg_arr;
   struct pvr_pds_ra_state ra_state;

   pvr_pds_ra_state_init(&ra_state);

   offset_reg_arr = pvr_pds_ra_allocate_64b_temp_contiguous(&ra_state, 1);
   total_reg_arr = pvr_pds_ra_allocate_64b_temp_contiguous(&ra_state, 1);
   ptemp_reg_arr = pvr_pds_ra_allocate_64b_temp_contiguous(&ra_state, 1);

   pvr_pds_ld_memory_region_setup(&ra_state,
                                  total_reg_arr,
                                  &total_count_addr_ld_setup_info);

   /* FIXME: Need to investigate where ptemps are loaded from external memory.
    */
   pvr_pds_ld_memory_region_setup(&ra_state,
                                  ptemp_reg_arr,
                                  &ptemp_ld_setup_info);

   pvr_pds_st_memory_region_setup(&ra_state,
                                  offset_reg_arr,
                                  &offset_count_addr_st_setup_info);

   if (gen_mode == PVR_PDS_GENERATE_CODE) {
      uint8_t *buffer = (uint8_t *)out->buffer;
      struct pvr_pds_instr instr;

      /* FIXME: Handle if (input->is_active) to use AddLD or AddMOV accordingly.
       */
      pvr_pds_build_stmc(&instr, PVR_PDS_SO_MASK_GLOBAL);
      buffer = pvr_pds_instr_encode_and_advance(&instr, buffer);

      /* LD Ptemp values from memory counter addresses. */
      buffer = pvr_pds_ld_memory_region_code(&ptemp_ld_setup_info, buffer);

      if (input->query_type == VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT) {
         /* TODO: Investigate if we need to clear the query result and
          * availability first.
          */
         buffer =
            pvr_pds_st_memory_region_code(&offset_count_addr_st_setup_info,
                                          buffer);

         /* TODO: After storing the current XFB total and fail counters to be
          * able to calculate the relative counts in the query's begin-end
          * scope, reset the query accumulated, total and failed.
          */
      }
   } else if (gen_mode == PVR_PDS_GENERATE_DATA) {
      uint32_t *buffer = (uint32_t *)out->buffer;

      pvr_pds_ld_memory_region_data(input->dev_info,
                                    &ptemp_ld_setup_info,
                                    input->result_addr,
                                    ptemp_reg_arr,
                                    buffer);

      pvr_pds_st_memory_region_data(input->dev_info,
                                    &offset_count_addr_st_setup_info,
                                    input->count_offset_addr,
                                    offset_reg_arr,
                                    buffer);

      pvr_pds_ld_memory_region_data(input->dev_info,
                                    &total_count_addr_ld_setup_info,
                                    input->total_count_addr,
                                    total_reg_arr,
                                    buffer);
   }

   /* Calculate total data size */
   out->data_size = pvr_pds_ra_get_allocated_32b_consts(&ra_state);

   out->code_size = offset_count_addr_st_setup_info.st_count4;
   out->code_size += total_count_addr_ld_setup_info.ld_count8;

   out->temp_size = ptemp_reg_arr.reg_count;
}
