/*
 * Copyright © 2024 Imagination Technologies Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef PVR_PDS_GEN_H
#define PVR_PDS_GEN_H

#include <stdbool.h>
#include <stdint.h>
#include "pvr_pds_ra.h"
#include "pvr_types.h"

enum pvr_pds_builder_generate_mode {
   PVR_PDS_GENERATE_CODE,
   PVR_PDS_GENERATE_DATA,
   PVR_PDS_GENERATE_SIZES,
};

/* TODO: derive counter size in bytes and dwords from device info features. */
#define PVR_XFB_COUNTER_SIZE_IN_DWORDS (1)

struct pvr_device_info;

struct pvr_pds_pixel_event_program_input {
   uint32_t emit_word_pair_count;

   /* Extra input required to generate the data segment. */
   struct {
      const struct pvr_device_info *dev_info;
      uint64_t task_control;
      const uint64_t *emit_words;
   } data;
};

struct pvr_pds_pixel_event_program_output {
   /* When generating the code or data segment. The buffer where to write out
    * the segment.
    */
   void *buffer;

   /* Always produced. The sizes of each program segment in dwords.
    */
   uint32_t data_size;
   uint32_t code_size;
};

enum pvr_pds_xfb_mode {
   /* All captured varying data goes into a single buffer. */
   PVR_PDS_TRANSFORM_FEEDBACK_MODE_INTERLEAVED,
   /* Each captured varying data goes into a separate buffer. */
   PVR_PDS_TRANSFORM_FEEDBACK_MODE_SEPARATE,
};

/* Corresponds to STREAM_OUT_BUFFER_INFO */
struct pvr_pds_xfb_buffer_info {
   pvr_dev_addr_t buffer_address;
   pvr_dev_addr_t buffer_size;
};

struct pvr_pds_xfb_varying_info {
   /* Varying size in dwords to write out. */
   uint32_t varying_size;
   /* Offset within the UVS to write out. */
   uint32_t input_offset;
   /* Offset at which to write out data into memory. */
   uint32_t output_offset;
};

struct pvr_pds_xfb_counter_buffer_info {
   uint32_t counter_buffer_varying_index;
};

struct pvr_pds_xfb_program_input {
   struct pvr_device_info *dev_info;

   enum pvr_pds_xfb_mode mode;

   /* Should the program generate overflow checking
      This is generally only needed in the following circumstances:
      - Robustness: For obvious reasons
      - Stream Queries: Where it's expected that it can fail and we need to
      track how many times that happens.
   */
   bool test_overflow;

   /* Transform feedback primitive type */
   uint16_t primitive_type;

   uint32_t varying_count;
   uint32_t varying_number[PVR_MAX_XFB_STREAMS];

   uint32_t persistent_temp_size;

   uint32_t counter_buffer_number;

   struct {
      pvr_dev_addr_t *counter_pairs_address;

      struct pvr_pds_xfb_varying_info *varying_info_arr;
      struct pvr_pds_xfb_buffer_info *buffer_info;

      struct pvr_pds_xfb_counter_buffer_info *counter_buffer_entries;

      /* Address of the buffer in external memory from where to load the ptemp
       * registers used in the main transform feedback program.
       */
      pvr_dev_addr_t ptemp_memory_buffer_addr;
   } data;
};
struct pvr_pds_xfb_init_program_input {
   bool load_from_counter_buffers;

   struct pvr_device_info *dev_info;

   uint16_t primitive_type;

   uint32_t persistent_temp_size;

   uint32_t counter_buffer_count;

   uint32_t num_loads;

   uint32_t *load_dest_select;

   struct {
      pvr_dev_addr_t *counter_pairs_address;
      pvr_dev_addr_t *counter_buffer_offsets_addr;

      struct pvr_pds_xfb_varying_info *varying_info_arr;
      struct pvr_pds_xfb_buffer_info *buffer_info;

      struct pvr_pds_xfb_counter_buffer_info *counter_buffer_entries;
      pvr_dev_addr_t *counter_buffers_addr;

      /* Address to the in memory buffer from where to load the ptemp registers
       * used in the main transform feedback program.
       */
      pvr_dev_addr_t ptemp_memory_buffer_addr;
   } data;
};

struct pvr_pds_xfb_end_program_input {
   bool store_counter_buffers;
   uint32_t buffer_strides[PVR_MAX_XFB_BUFFERS];
   uint32_t counter_buffer_count;

   struct pvr_device_info *dev_info;

   struct {
      pvr_dev_addr_t *counter_buffers_addr;

      pvr_dev_addr_t *counter_pairs_address;
      pvr_dev_addr_t *counter_buffer_offsets_addr;
   } data;
};

struct pvr_pds_xfb_draw_program_input {
   uint32_t first_instance;
   uint32_t instance_count;
   uint32_t vertex_stride;

   struct pvr_device_info *dev_info;
   uint32_t negative_counter_offset;
   struct {
      /* TODO: Should Address be 128 bits aligned, and 32bits size. */
      pvr_dev_addr_t counter_buffer_addr;
      pvr_dev_addr_t ouput_count_buffer_addr;
   } data;
};

struct pvr_pds_xfb_query_program_input {
   struct pvr_device_info *dev_info;

   VkQueryType query_type;
   uint32_t counter_bytes;

   struct {
      /* True if transform feedback is active. */
      bool is_active : 1;
      /* True if generating for CmdEndxQueryIndexedEXT. */
      bool is_store : 1;
      /* True if generating for CmdBeginQueryIndexedEXT. */
      bool is_init : 1;
   };

   pvr_dev_addr_t result_addr;
   pvr_dev_addr_t availability_addr;

   pvr_dev_addr_t count_offset_addr;
   pvr_dev_addr_t count_offset_failed_addr;
   pvr_dev_addr_t count_offset_accumulated_addr;

   /* Only set when is_active is true. */
   pvr_dev_addr_t total_count_addr;
};

struct pvr_pds_xfb_program_output {
   void *buffer;

   /* The sizes of each program segment in dwords */
   uint32_t data_size;
   uint32_t code_size;
   uint32_t temp_size;
};

void pvr_pds_gen_pixel_event_program(
   enum pvr_pds_builder_generate_mode gen_mode,
   const struct pvr_pds_pixel_event_program_input *input,
   struct pvr_pds_pixel_event_program_output *out);

void pvr_pds_gen_xfb_program(enum pvr_pds_builder_generate_mode gen_mode,
                             const struct pvr_pds_xfb_program_input *input,
                             struct pvr_pds_xfb_program_output *out);

void pvr_pds_gen_xfb_init_program(
   enum pvr_pds_builder_generate_mode gen_mode,
   const struct pvr_pds_xfb_init_program_input *input,
   struct pvr_pds_xfb_program_output *out);

void pvr_pds_gen_transform_end_program(
   enum pvr_pds_builder_generate_mode gen_mode,
   const struct pvr_pds_xfb_end_program_input *input,
   struct pvr_pds_xfb_program_output *out);

void pvr_pds_gen_generate_vertex_count_program(
   enum pvr_pds_builder_generate_mode gen_mode,
   const struct pvr_pds_xfb_draw_program_input *input,
   struct pvr_pds_xfb_program_output *out);

void pvr_pds_gen_generate_query_program(
   enum pvr_pds_builder_generate_mode gen_mode,
   const struct pvr_pds_xfb_query_program_input *input,
   struct pvr_pds_xfb_program_output *out);

#endif /* PVR_PDS_GEN_H */
