/*
 * Copyright © 2024 Imagination Technologies Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <assert.h>
#include <stdint.h>

#include "pds/pvr_pds_isa.h"
#include "pvr_pds_ra.h"
#include "util/bitset.h"

struct pvr_pds_reg_array
pvr_pds_ra_allocate_32b_const_contiguous(struct pvr_pds_ra_state *state,
                                         uint32_t count)
{
   uint32_t const_idx;

   if (count == 1 && !BITSET_IS_EMPTY(state->skipped_consts)) {
      const_idx = BITSET_FFS(state->skipped_consts);

      BITSET_CLEAR(state->skipped_consts, const_idx);
   } else {
      const_idx = state->next_free_const;

      assert(const_idx + count <= 128);
      state->next_free_const += count;
   }

   return (struct pvr_pds_reg_array){
      .reg0 = {
         .type = PVR_PDS_REG_TYPE_CONST,
         .reg_size = PVR_PDS_REG_SIZE_32,
         .idx = const_idx,
      },
      .reg_count = count,
   };
}

struct pvr_pds_reg_array
pvr_pds_ra_allocate_64b_const_contiguous(struct pvr_pds_ra_state *state,
                                         uint32_t count)
{
   /* Each const reg is 32 bits wide, so we need to allocate 2 registers at
    * a 64 bit boundary for each requested 64b reg.
    */
   uint32_t const_idx = ALIGN_POT(state->next_free_const, 2);

   assert(const_idx + 2 * count <= 128);

   /* If we're skipping over a 32b const for the 64b alignment. */
   if (state->next_free_const & 1)
      BITSET_SET(state->skipped_consts, state->next_free_const);

   state->next_free_const = const_idx + 2 * count;

   assert(const_idx != -1);
   assert((const_idx & 1) == 0);

   return (struct pvr_pds_reg_array){
      .reg0 = {
         .type = PVR_PDS_REG_TYPE_CONST,
         .reg_size = PVR_PDS_REG_SIZE_64,
         .idx = const_idx / 2,
      },
      .reg_count = count,
   };
}

static uint32_t
pvr_pds_ra_bitmap_allocate_contiguos(uint32_t *bitmap,
                                     enum pvr_pds_reg_size reg_size,
                                     uint32_t count)
{
   uint32_t availability = *bitmap;
   uint32_t shifted_mask = 0;
   uint32_t reg_idx = ~0;
   uint32_t mask;

   /* Check that we haven't run out of regs. */
   assert(*bitmap != 0);

   if (reg_size == PVR_PDS_REG_SIZE_64)
      count *= 2;

   mask = BITFIELD_MASK(count);

   while (availability != 0) {
      uint32_t bit;

      static_assert(sizeof(availability) == sizeof(*bitmap),
                    "mismatched size.");
      static_assert(sizeof(*bitmap) == sizeof(int), "ffs take in an int.");
      bit = ffs(availability) - 1;

      if (reg_size == PVR_PDS_REG_SIZE_64 && bit % 2 == 1) {
         /* We need a 64b aligned reg. */
         availability &= BITFIELD_BIT(bit);
         continue;
      }

      assert(bit <= 32 - count);

      shifted_mask = mask << bit;

      if ((availability & shifted_mask) == shifted_mask) {
         reg_idx = bit;
         break;
      }

      availability &= ~shifted_mask;
   }

   assert(reg_idx != ~0);

   *bitmap &= ~shifted_mask;

   return reg_idx;
}

struct pvr_pds_reg pvr_pds_ra_reserve_32b_temp(struct pvr_pds_ra_state *state,
                                               enum pvr_pds_prefilled_temp idx)
{
   assert((state->temp_bitmap & BITSET_BIT(idx)) != 0);

   state->temp_bitmap &= ~BITSET_BIT(idx);

   return (struct pvr_pds_reg){
      .type = PVR_PDS_REG_TYPE_TEMP,
      .reg_size = PVR_PDS_REG_SIZE_32,
      .idx = idx,
   };
}

struct pvr_pds_reg_array
pvr_pds_ra_allocate_32b_temp_contiguous(struct pvr_pds_ra_state *state,
                                        uint32_t count)
{
   const uint32_t reg_idx =
      pvr_pds_ra_bitmap_allocate_contiguos(&state->temp_bitmap,
                                           PVR_PDS_REG_SIZE_32,
                                           count);

   return (struct pvr_pds_reg_array){
      .reg0 = {
         .type = PVR_PDS_REG_TYPE_TEMP,
         .reg_size = PVR_PDS_REG_SIZE_32,
         .idx = reg_idx,
      },
      .reg_count = count,
   };
}

struct pvr_pds_reg_array
pvr_pds_ra_allocate_64b_temp_contiguous(struct pvr_pds_ra_state *state,
                                        uint32_t count)
{
   const uint32_t reg_idx =
      pvr_pds_ra_bitmap_allocate_contiguos(&state->temp_bitmap,
                                           PVR_PDS_REG_SIZE_64,
                                           count);

   return (struct pvr_pds_reg_array){
      .reg0 = {
         .type = PVR_PDS_REG_TYPE_TEMP,
         .reg_size = PVR_PDS_REG_SIZE_64,
         .idx = reg_idx,
      },
      .reg_count = count, 
   };
}

struct pvr_pds_reg_array pvr_pds_ra_allocate_32b_persistent_temp_contiguous(
   struct pvr_pds_ra_state *state,
   uint32_t count)
{
   const uint32_t reg_idx =
      pvr_pds_ra_bitmap_allocate_contiguos(&state->persistent_temp_bitmap,
                                           PVR_PDS_REG_SIZE_32,
                                           count);

   return (struct pvr_pds_reg_array){
      .reg0 = {
         .type = PVR_PDS_REG_TYPE_PTEMP,
         .reg_size = PVR_PDS_REG_SIZE_32,
         .idx = reg_idx,
      },
      .reg_count = count,
   };
}

struct pvr_pds_reg_array pvr_pds_ra_allocate_64b_persistent_temp_contiguous(
   struct pvr_pds_ra_state *state,
   uint32_t count)
{
   const uint32_t reg_idx =
      pvr_pds_ra_bitmap_allocate_contiguos(&state->persistent_temp_bitmap,
                                           PVR_PDS_REG_SIZE_64,
                                           count);

   return (struct pvr_pds_reg_array){
      .reg0 = {
         .type = PVR_PDS_REG_TYPE_PTEMP,
         .reg_size = PVR_PDS_REG_SIZE_64,
         .idx = reg_idx,
      },
      .reg_count = count,
   };
}
