/*
 * Copyright © 2024 Imagination Technologies Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef PVR_PDS_RA_H
#define PVR_PDS_RA_H

#include <stdint.h>

#include "pds/pvr_pds_isa.h"
#include "util/bitset.h"

struct pvr_pds_ra_state {
   uint32_t next_free_const;
   /* When allocating 64 bit consts, we can waste a 32 bit const due to the
    * 64 bit alignment requirement. We keep track of those holes with this
    * bitmap, so that we can attempt to reuse them on the next 32 bit
    * allocation.
    *
    * A bit 1 indicates a 32 bit const is available for reuse.
    */
   BITSET_WORD skipped_consts[BITSET_WORDS(PVR_PDS_REG_CONST_COUNT)];

   uint32_t temp_bitmap;
   static_assert(PVR_PDS_REG_TEMP_COUNT == 32,
                 "temp_bitmap doesn't cover all temp regs");

   uint32_t persistent_temp_bitmap;
   static_assert(PVR_PDS_REG_PTEMP_COUNT == 32,
                 "persistent_temp_bitmap doesn't cover all ptemp regs");
};

struct pvr_pds_reg_array {
   struct pvr_pds_reg reg0;
   uint32_t reg_count;
};

static void pvr_pds_ra_state_init(struct pvr_pds_ra_state *state)
{
   *state = (struct pvr_pds_ra_state){
      .temp_bitmap = ~0,
      .persistent_temp_bitmap = ~0,
   };
}

static inline struct pvr_pds_reg
pvr_pds_reg_from_array(const struct pvr_pds_reg_array array, uint32_t idx)
{
   struct pvr_pds_reg reg = array.reg0;

   assert(idx < array.reg_count);

   reg.idx += idx;

   return reg;
}

struct pvr_pds_reg_array
pvr_pds_ra_allocate_32b_const_contiguous(struct pvr_pds_ra_state *state,
                                         uint32_t count);
struct pvr_pds_reg_array
pvr_pds_ra_allocate_64b_const_contiguous(struct pvr_pds_ra_state *state,
                                         uint32_t count);

static struct pvr_pds_reg
pvr_pds_ra_allocate_32b_const(struct pvr_pds_ra_state *state)
{
   const struct pvr_pds_reg_array reg_arr =
      pvr_pds_ra_allocate_32b_const_contiguous(state, 1);

   return pvr_pds_reg_from_array(reg_arr, 0);
}

static struct pvr_pds_reg
pvr_pds_ra_allocate_64b_const(struct pvr_pds_ra_state *state)
{
   const struct pvr_pds_reg_array reg_arr =
      pvr_pds_ra_allocate_64b_const_contiguous(state, 1);

   return pvr_pds_reg_from_array(reg_arr, 0);
}

static uint32_t
pvr_pds_ra_get_allocated_32b_consts(const struct pvr_pds_ra_state *state)
{
   return state->next_free_const;
}

/**
 * \brief Reserves a pre-filled temporary register.
 *
 * The register must be free (i.e. not allocated previously). Reservations
 * should be made prior to any temp allocations as to reduce the chance of such
 * conflict.
 *
 * Reserving a temp is necessary if the program intends to make use of a
 * pre-filled temp register's contents as to avoid them from being allocated
 * for other purposes within the program.
 */
struct pvr_pds_reg pvr_pds_ra_reserve_32b_temp(struct pvr_pds_ra_state *state,
                                               enum pvr_pds_prefilled_temp idx);

struct pvr_pds_reg_array
pvr_pds_ra_allocate_32b_temp_contiguous(struct pvr_pds_ra_state *state,
                                        uint32_t count);
struct pvr_pds_reg_array
pvr_pds_ra_allocate_64b_temp_contiguous(struct pvr_pds_ra_state *state,
                                        uint32_t count);

static struct pvr_pds_reg
pvr_pds_ra_allocate_32b_temp(struct pvr_pds_ra_state *state)
{
   const struct pvr_pds_reg_array reg_arr =
      pvr_pds_ra_allocate_32b_temp_contiguous(state, 1);

   return pvr_pds_reg_from_array(reg_arr, 0);
}

static struct pvr_pds_reg
pvr_pds_ra_allocate_64b_temp(struct pvr_pds_ra_state *state)
{
   const struct pvr_pds_reg_array reg_arr =
      pvr_pds_ra_allocate_64b_temp_contiguous(state, 1);

   return pvr_pds_reg_from_array(reg_arr, 0);
}

struct pvr_pds_reg_array pvr_pds_ra_allocate_32b_persistent_temp_contiguous(
   struct pvr_pds_ra_state *state,
   uint32_t count);
struct pvr_pds_reg_array pvr_pds_ra_allocate_64b_persistent_temp_contiguous(
   struct pvr_pds_ra_state *state,
   uint32_t count);

static struct pvr_pds_reg
pvr_pds_ra_allocate_32b_persistent_temp(struct pvr_pds_ra_state *state)
{
   const struct pvr_pds_reg_array reg_arr =
      pvr_pds_ra_allocate_32b_persistent_temp_contiguous(state, 1);

   return pvr_pds_reg_from_array(reg_arr, 0);
}

static struct pvr_pds_reg
pvr_pds_ra_allocate_64b_persistent_temp(struct pvr_pds_ra_state *state)
{
   const struct pvr_pds_reg_array reg_arr =
      pvr_pds_ra_allocate_64b_persistent_temp_contiguous(state, 1);

   return pvr_pds_reg_from_array(reg_arr, 0);
}

static inline struct pvr_pds_reg
pvr_pds_ra_low_32b_from_64b(const struct pvr_pds_reg reg)
{
   assert(reg.reg_size == PVR_PDS_REG_SIZE_64);
   // assert(reg.idx % 2 == 0);
   // FIXME: Why is this a requierment here but not while allocating the registers

   return (struct pvr_pds_reg){
      .type = reg.type,
      .reg_size = PVR_PDS_REG_SIZE_32,
      .idx = reg.idx * 2,
   };
}

static inline struct pvr_pds_reg
pvr_pds_ra_high_32b_from_64b(const struct pvr_pds_reg reg)
{
   assert(reg.reg_size == PVR_PDS_REG_SIZE_64);
   // assert(reg.idx % 2 == 0);
   // FIXME: Why is this a requierment here but not while allocating the registers
   return (struct pvr_pds_reg){
      .type = reg.type,
      .reg_size = PVR_PDS_REG_SIZE_32,
      .idx = reg.idx * 2 + 1,
   };
}

#endif /* PVR_PDS_RA_H */
